Rightist lists aka standard cons lists.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.List where

open import VFP.Boolean
open import VFP.Maybe hiding (map ; return ; _>>=_)
open import VFP.Natural.Order
open import VFP.Decidable hiding ([_])
open import VFP.Equality

private
  variable
    A Elem : Set

infixr 7 _⧺_
infix  6 _∈′_
\end{code}

-------------------------------------------------------------------------------

<<cons-fixity>>
\begin{code}
infixr 7 _∷_
\end{code}

<<List>>
\begin{code}
data List (Elem : Set) : Set where
  []   : List Elem
  _∷_  : Elem -> List Elem -> List Elem
\end{code}

\begin{code}
𝕃 = List
\end{code}

<<first-five-primes>>
\begin{code}
first-five-primes  :  List ℕ
first-five-primes  =  2 ∷ (3 ∷ (5 ∷ (7 ∷ (11 ∷ []))))
\end{code}

\begin{code}
{-# BUILTIN LIST List #-}
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
null : List A -> Boolean
null []       = true
null (a ∷ as) = false
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<map>>
\begin{code}
map : ∀ {A B : Set} -> (A -> B) -> (List A -> List B)
map f []        = []
map f (x ∷ xs)  = f x ∷ map f xs
\end{code}

\begin{code}
infix 8 [_]

[_] : ∀ {Elem : Set} -> Elem → List Elem
[ x ] = x ∷ []

return : ∀ {A : Set} -> A -> List A
return a = a ∷ []
\end{code}

<<append>>
\begin{code}
_⧺_ : {Elem : Set} -> List Elem → List Elem → List Elem
[]        ⧺ ys = ys
(x ∷ xs)  ⧺ ys = x ∷ (xs ⧺ ys)
\end{code}

\begin{code}
concat : ∀ {A : Set} -> List (List A) -> List A
concat []         = []
concat (as ∷ aas) = as ⧺ concat aas
\end{code}

<<length-with-type-argument>>
\begin{code}
length′ : (Elem : Set) → List Elem → ℕ
length′ Elem []        = zero
length′ Elem (x ∷ xs)  = succ (length′ Elem xs)
\end{code}

<<length-fun>>
\begin{code}
length-fun  : List (ℕ → ℕ) → ℕ
length-fun  = length′ (ℕ → ℕ)
\end{code}

<<length>>
\begin{code}
length : {Elem : Set} → List Elem → ℕ
length []        = zero
length (x ∷ xs)  = succ (length xs)
\end{code}

length []       = 0
length (x ∷ xs) = 1 + length xs

<<snoc>>
\begin{code}
xsnoc : {Elem : Set} -> List Elem -> Elem -> List Elem
xsnoc []        y  = y ∷ []
xsnoc (x ∷ xs)  y  = x ∷ xsnoc xs y
\end{code}

<<foldr>>
\begin{code}
foldr : ∀ {P S : Set} -> (P -> S -> S) -> S -> List P -> S
foldr _◁_ s []        = s
foldr _◁_ s (x ∷ xs)  = x ◁ (foldr _◁_ s xs)
\end{code}

<<foldr-simple>>
\begin{code}
foldr-simple : ∀ {Elem Result : Set} ->
  (Elem -> Result -> Result) -> Result -> List Elem -> Result
foldr-simple c n []        = n
foldr-simple c n (x ∷ xs)  = c x (foldr-simple c n xs)
\end{code}


\begin{code}
foldl : {P S : Set} -> (S -> P -> S) -> S -> List P -> S
foldl _▷_ s []        = s
foldl _▷_ s (x ∷ xs)  = foldl _▷_ (s ▷ x) xs
\end{code}

<<sum-directly>>
\begin{code}
sum :  List ℕ -> ℕ
sum []        = 0
sum (x ∷ xs)  = x + sum xs 
\end{code}

<<sum-primes>>
\begin{code}
sum-primes : ℕ
sum-primes = sum first-five-primes
\end{code}

<<length-primes>>
\begin{code}
length-primes : ℕ
length-primes = length′ ℕ first-five-primes
\end{code}

<<double-primes>>
\begin{code}
double-primes : List ℕ
double-primes = map double first-five-primes
\end{code}

<<sum>>
\begin{code}
sum'  :  List ℕ -> ℕ
sum'  =  foldr _+_ 0
\end{code}

\begin{code}
and or : List 𝔹 -> 𝔹
and = foldr _&&_ true
or  = foldr _||_ false
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
from-of-length : ℕ → ℕ → List ℕ
from-of-length l (zero)    =  []
from-of-length l (succ n)  =  l ∷ from-of-length (1 + l) n

[_‥_] : ℕ → ℕ → List ℕ
[ l ‥ r ] = from-of-length l (r ∸ l)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
mapMaybe : ∀ {A B : Set} -> (A -> Maybe B) -> (List A -> List B)
mapMaybe f [] = []
mapMaybe f (a ∷ as) with f a
... | Nothing =     mapMaybe f as
... | Just b  = b ∷ mapMaybe f as
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
filter : ∀ {P : A -> Set} → Decidable P → List A → List A
filter P? [] = []
filter P? (x ∷ xs) with P? x
... | yes p = x ∷ filter P? xs
... | no ¬p =     filter P? xs
\end{code}

-------------------------------------------------------------------------------

<<reverse>>
\begin{code}
reverse : {Elem : Set} → List Elem → List Elem
reverse []        = []
reverse (x ∷ xs)  = reverse xs ⧺ [ x ]
\end{code}


<<reverse-book>>
\begin{code}
reverse-book : {Elem : Set} → List Elem → List Elem
reverse-book []        = []
reverse-book (x ∷ xs)  = xsnoc (reverse-book xs) x
\end{code}


reverse (x ∷ xs)  = snoc (reverse xs) x

<<slow-reverse>>
\begin{code}
slow-reverse = reverse
\end{code}


<<slow-reverse-book>>
\begin{code}
slow-reverse-book = reverse-book
\end{code}

<<fast-reverse>>
\begin{code}
reverse-append : {Elem : Set} → List Elem → List Elem → List Elem
reverse-append []        ys  = ys
reverse-append (x ∷ xs)  ys  = reverse-append xs (x ∷ ys)

fast-reverse : {Elem : Set} → List Elem → List Elem
fast-reverse xs = reverse-append xs []
\end{code}

\begin{code}
lookup : (xs : List Elem) -> {n : ℕ} -> n ≺ length xs -> Elem
lookup (x ∷ xs) (zero≺succ)   = x
lookup (x ∷ xs) (succ≺succ p) = lookup xs p
\end{code}

-------------------------------------------------------------------------------

Todo: Move all and exist into new subfile List.Quantifier

Predicates i.e. set-valued functions. (forall is a keyword, which is
why the universal predicate is called all not forall. Haskell defines
all , any :: (a -> Bool) -> ([a] -> Bool).)

<<all>>
\begin{code}
data all (P : Elem -> Set) : (List Elem -> Set) where
  []   :                                 all P []
  _∷_  : ∀ {a as} -> P a -> all P as ->  all P (a ∷ as)
\end{code}

\begin{code}
zip : {P : Elem -> Set} -> (xs : List Elem) -> all P xs -> List (Σ Elem P)
zip []       []       = []
zip (x ∷ xs) (p ∷ ps) = (x , p) ∷ zip xs ps
\end{code}

NB. The constructors [] and _∷_ are overloaded.

The empty list trivially satisfies the property |all P|; a non-empty
list satisfies the property |all P| if the head and each element of
the tail satisfies the property |P|.

<<exists>>
\begin{code}
data exists (P : Elem -> Set) : (List Elem -> Set) where
  head  : ∀ {a as} -> P a          -> exists P (a ∷ as)
  tail  : ∀ {a as} -> exists P as  -> exists P (a ∷ as)
\end{code}

A non-empty list satisfies the property |exists P| if either the head
satisfies |P| or any element in the tail (the constructors head and
tail can be seen as paths).

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Membership.

begin{code}
data _∈_ (x : A) : List A -> Set where
  head : ∀ {as} → x ∈ x ∷ as
  tail : ∀ {a as} → x ∈ as → x ∈ a ∷ as 
end{code}

<<membership>>
\begin{code}
_∈′_ : A -> List A -> Set
a ∈′ xs = exists (λ x -> a ≡ x) xs
\end{code}

-------------------------------------------------------------------------------
