Six ways to define a strict ordering on the natural numbers.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural.Order where

open import VFP.Natural public
open import VFP.Equality public
open import VFP.Logic public

private
  variable
    k m n : ℕ

infix 6 _≺_
infix 6 _≺₇_ _≼₇_ _≻₇_  _≽₇_
\end{code}

-------------------------------------------------------------------------------

Strict order.

<<strict-order>>
\begin{code}
data _≺_ : Natural → Natural → Set where
  zero≺succ :         zero   ≺ succ n
  succ≺succ : m ≺ n → succ m ≺ succ n
\end{code}

\begin{code}
_ :  3 ≺ 5
_ =  succ≺succ (succ≺succ (succ≺succ zero≺succ))
\end{code}

In words, |zero| is the smallest element; |succ| is strictly
monotone. This is an example of a lexicographic ordering: the
constructor |zero| is smaller than the constructor |succ|; if the
constructors are the same, the arguments are related by |_≺_|.

<<strict-reverse-order>>
\begin{code}
infix 6 _≻_
_≻_ : ℕ → ℕ → Set
m ≻ n = n ≺ m
\end{code}

Derived non-strict order.

<<non-strict-order>>
\begin{code}
infix 6 _≼_ _≽_

_≼_ : ℕ -> ℕ -> Set
m ≼ n = m ≡ n ⊎ m ≺ n

_≽_ : ℕ -> ℕ -> Set
m ≽ n = n ≼ m
\end{code}

Same as above, except that the constructors are more speaking:
inl reflexive is EQ and inr p is LT p.

\begin{code}
infix 6 _⩽_ _⩾_

data _⩽_ : Natural → Natural → Set where
  EQ : n ⩽ n
  LT : m ≺ n -> m ⩽ n

_⩾_ : ℕ -> ℕ -> Set
m ⩾ n = n ⩽ m
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

A relation R is called connected when x R y or x ≡ y or y R x. A
relation with the property that x R y or y R x is called strongly
connected. Trichotomy insists that that exactly one of the three
options x R y , x ≡ y, and y R x holds.

<<Ordering-compare>>
\begin{code}
data Ordering : ℕ → ℕ → Set where
  LT  : {ev : m ≺ n}  → Ordering m n
  EQ  : {ev : m ≡ n}  → Ordering m n
  GT  : {ev : m ≻ n}  → Ordering m n

compare : ∀ m n → Ordering m n
compare (zero)    (zero)    = EQ {ev = reflexive}
compare (zero)    (succ n)  = LT {ev = zero≺succ}
compare (succ m)  (zero)    = GT {ev = zero≺succ}
compare (succ m)  (succ n) with compare m n
... | LT {ev = m≺n}         = LT {ev = succ≺succ m≺n}
... | EQ {ev = reflexive}   = EQ {ev = reflexive}
... | GT {ev = m≻n}         = GT {ev = succ≺succ m≻n}
\end{code}

OR, ditch equality proof:

  EQ : Ordering m m

\begin{code}
trichotomy : ∀ m n -> m ≺ n ⊎ m ≡ n ⊎ m ≻ n
trichotomy m n with compare m n
... | LT {ev = m≺n} = inl m≺n
... | EQ {ev = m≡n} = inr (inl m≡n)
... | GT {ev = m≻n} = inr (inr m≻n)
\end{code}

Same as compare, using a different, yet isomorphic result type.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Order with class.

\begin{code}
data Ordering′ : ℕ → ℕ → Set where
  LT : ⦃ ev : m ≺ n ⦄ → Ordering′ m n
  EQ :                  Ordering′ m m
  GT : ⦃ ev : m ≻ n ⦄ → Ordering′ m n

compare′ : ∀ m n → Ordering′ m n
compare′ m n with compare m n
... | LT {ev = ev}        = LT ⦃ ev = ev ⦄
... | EQ {ev = reflexive} = EQ
... | GT {ev = ev}        = GT ⦃ ev = ev ⦄
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Relating order and arithmetic.

<<order-and-arithmetic>>
\begin{code}
n≺succ-n : ∀ {n} → n ≺ succ n
n≺succ-n {zero}    = zero≺succ
n≺succ-n {succ n}  = succ≺succ n≺succ-n
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

The strict order is well-founded.

begin{code}
data Accessible (n : ℕ) : Set
Accessible≺ : ℕ → Set

data Accessible n where
  access : Accessible≺ n → Accessible n

Accessible≺ n = ∀ {m} → m ≺ n → Accessible m
end{code}

<<Accessible>>
\begin{code}
data Accessible (n : ℕ) : Set where
  access : (∀ {m} → m ≺ n → Accessible m) → Accessible n
\end{code}

\begin{code}
Accessible≺ : ℕ → Set
Accessible≺ n = ∀ {m} → m ≺ n → Accessible m
\end{code}

\begin{code}
acc-zero : Accessible≺ zero
acc-zero = λ ()

acc-succ : ∀ {n} → Accessible n → Accessible≺ (succ n)
acc-succ (access hyp) (zero≺succ)     = access acc-zero
acc-succ (access hyp) (succ≺succ m≺n) = access (acc-succ (hyp m≺n))

well-founded : ∀ {n} → Accessible n
well-founded {zero}    = access acc-zero
well-founded {succ n}  = access (acc-succ well-founded)

accessible = well-founded
\end{code}

NB. Observe that the "access" constructor is generated
immediately. The old version, shown below, didn't do this, resulting
in a bad rewriting behaviour (see Natural.Lg).

\begin{code}
acc-zero′ : Accessible zero
acc-zero′ = access λ ()

acc-succ′ : Accessible n → Accessible (succ n)
acc-succ′ (access hyp) = access λ where
  (zero≺succ)     → acc-zero′
  (succ≺succ m≺n) → acc-succ′ (hyp m≺n)

well-founded′ : ∀ {n} → Accessible n
well-founded′ {zero}    = acc-zero′
well-founded′ {succ n}  = acc-succ′ (well-founded′ {n})
\end{code}

|n≺succ-n| shows that structural induction/recursion is a special case
of strong/Noetherian induction/recursion.  The definition of
|acc-succ| uses an anonymous pattern matching function.

Typical pattern of |Accessible| recursion: |f (access because)| triggers a
recursive call of the form |f (because proof)|.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Minimum and maximum, see also VFP.Natural.

\begin{code}
infixr 7 _↓_ _↑_

_↓_ _↑_ : Natural -> Natural -> Natural
a ↓ b with compare a b
... | LT = a
... | EQ = a
... | GT = b
a ↑ b with compare a b
... | LT = b
... | EQ = b
... | GT = a
\end{code}

-------------------------------------------------------------------------------

Alternative definition of the strict order.

<<strict-order-1>>
\begin{code}
infix 6 _≺₁_
data _≺₁_ : ℕ → ℕ → Set where
  one   :           n ≺₁ succ n  -- difference is exactly |1|
  succ  : m ≺₁ n →  m ≺₁ succ n  -- increase right bound by |1|

_ :  3 ≺₁ 5
_ =  succ one
\end{code}

\begin{code}
infix 6 _≻₁_

_≻₁_ : ℕ → ℕ → Set
m ≻₁ n = n ≺₁ m
\end{code}

\begin{code}
data Accessible₁ (n : ℕ) : Set where
  access : (∀ {m} → m ≺₁ n → Accessible₁ m) → Accessible₁ n

acc-zero₁ : Accessible₁ zero
acc-zero₁ = access λ ()

acc-succ₁ : Accessible₁ n → Accessible₁ (succ n)
acc-succ₁ access-n@(access hyp) = access λ where
  (one)       → access-n
  (succ m≺₁n) → hyp m≺₁n

well-founded₁ : ∀ n → Accessible₁ n
well-founded₁ (zero)    = acc-zero₁
well-founded₁ (succ n)  = acc-succ₁ (well-founded₁ n)
\end{code}

-------------------------------------------------------------------------------

Exercise: symmetric version, decrease left bound.


  -- difference is exactly |1|
  -- decrease left bound by |1|
  
<<strict-order-7>>
\begin{code}
data _≺₇_ : ℕ → ℕ → Set where
  one   : n ≺₇ 1 + n
  succ  : 1 + m ≺₇ n →  m ≺₇ n
\end{code}

<<non-strict-order-7>>
\begin{code}
_≼₇_ : ℕ -> ℕ -> Set
m ≼₇ n = (m ≡ n) ⊎ (m ≺₇ n)
\end{code}

\begin{code}
_≻₇_ : ℕ → ℕ → Set
m ≻₇ n = n ≺₇ m

_≽₇_ : ℕ -> ℕ -> Set
m ≽₇ n = n ≼₇ m
\end{code}

begin{code}
_ : 4 ≺₇ 7
_ = succ (succ one)
end{code}

<<order-2>>
\begin{code}
data _≼₂_ : ℕ → ℕ → Set where
  zero  : n ≼₂ n                 -- difference is exactly |0|
  succ  : succ m ≼₂ n →  m ≼₂ n  -- decrease left bound by |1|
\end{code}

\begin{code}
_≺₂_ : ℕ → ℕ → Set
m ≺₂ n = succ m ≼₂ n
\end{code}

-------------------------------------------------------------------------------

Or: use n ≺₁ succ n and transitivity (proofs are no longer unique).

\begin{code}
data _≺₃_ : ℕ → ℕ → Set where
  one           : n ≺₃ succ n               -- one-point interval
  ≺₃-transitive : k ≺₃ m → m ≺₃ n → k ≺₃ n  -- transitivity: add intervals
\end{code}

TODO Accessible₅.

-------------------------------------------------------------------------------

A generic approach: works for *any* natural number datatype that
supports addition (the so-called algebraic ordering).

<<strict-order-4>>
\begin{code}
open import VFP.Logic

infix 6 _≺₄_ _≻₄_

_≺₄_ : ℕ -> ℕ -> Set
m ≺₄ n = ∃ (λ Δ -> succ (Δ + m) ≡ n)

_≻₄_ : ℕ → ℕ → Set
m ≻₄ n = n ≺₄ m
\end{code}

--m ≺₄ n = Σ ℕ (λ Δ -> succ (Δ + m) ≡ n)

TODO Accessible₄.

data Accessible (n : ℕ) : Set where
  access : (∀ {m} → m ≺ n → Accessible m) → Accessible n

Accessible≺ : ℕ → Set
Accessible≺ n = ∀ {m} → m ≺ n → Accessible m


acc-zero : Accessible≺ zero
acc-zero = λ ()

acc-succ : ∀ {n} → Accessible n → Accessible≺ (succ n)
acc-succ (access hyp) (zero≺succ)     = access acc-zero
acc-succ (access hyp) (succ≺succ m≺n) = access (acc-succ (hyp m≺n))

well-founded : ∀ {n} → Accessible n
well-founded {zero}    = access acc-zero
well-founded {succ n}  = access (acc-succ well-founded)

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Same as above, using a tailormade datatype (slightly nicer to use).

\begin{code}
data _≺₅_ : ℕ → ℕ → Set where
  diff : (Δ : ℕ) -> n ≺₅ succ (Δ + n)            -- difference is exactly |succ Δ|
\end{code}

-------------------------------------------------------------------------------

Exercise: details matter! Repeat the development using the following
version where Δ and m are flipped.

\begin{code}
infix 6 _≺₆_ _≻₆_

_≺₆_ : ℕ -> ℕ -> Set
m ≺₆ n = Σ ℕ (λ Δ -> succ (m + Δ) ≡ n)

_≻₆_ : ℕ → ℕ → Set
m ≻₆ n = n ≺₆ m
\end{code}

\begin{code}
open import VFP.Natural.Properties

data Accessible₆ (n : ℕ) : Set where
  access : (∀ {m} → m ≺₆ n → Accessible₆ m) → Accessible₆ n

acc-zero₆ : Accessible₆ zero
acc-zero₆ = access λ ()

acc-succ₆ : ∀ {n : ℕ} -> Accessible₆ n → Accessible₆ (succ n)
acc-succ₆ {n} (access hyp) = access hyp′
  where hyp′ : {m : ℕ} → m ≺₆ succ n → Accessible₆ m
        hyp′ {m} (zero , reflexive) rewrite +-right-identity {m} = access hyp
        hyp′ (succ Δ , p) = hyp (Δ , succ-injective (transitive (congruent succ (shift' _ _)) p))

well-founded₆ : ∀ n → Accessible₆ n
well-founded₆ (zero)    = acc-zero₆
well-founded₆ (succ n)  = acc-succ₆ (well-founded₆ n)
\end{code}

-------------------------------------------------------------------------------

Unicode: ≺ \prec ; ≻ \succ ; ≼ \preceq ; ≽ \succeq
