\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural.Properties where

open import VFP.Natural public
open import VFP.Equality.Properties public
open import VFP.Equality.Reasoning
open import VFP.Logic
open import VFP.Decidable

private
  variable
    k m n : ℕ
\end{code}

-------------------------------------------------------------------------------

Data constructors are injective.

<<succ-injective>>
\begin{code}
succ-injective : ∀ {m n : ℕ} -> succ m ≡ succ n -> m ≡ n
succ-injective reflexive = reflexive
\end{code}

No junk, no confusion.

<<no-junk>>
\begin{code}
no-junk : ∀ (n : ℕ) -> n ≡ zero ⊎ ∃ (λ m -> n ≡ succ m)
no-junk (zero)    = inl  reflexive
no-junk (succ n)  = inr  (n , reflexive)
\end{code}

<<no-confusion>>
\begin{code}
no-confusion : ¬ ∃ (λ (n : ℕ) -> zero ≡ succ n)
no-confusion ()
\end{code}

\begin{code}
no-confusion′ : ∀ (n : ℕ) -> ¬ zero ≡ succ n
no-confusion′ (zero)   ()
no-confusion′ (succ n) ()
\end{code}

Decide equality.

\begin{code}
infix 6 _≡?_  

_≡?_ : (m n : ℕ) → Decide (m ≡ n)
zero   ≡? zero   = yes reflexive
zero   ≡? succ m = no (λ ())
succ n ≡? zero   = no (λ ())
succ n ≡? succ m with n ≡? m
… | yes reflexive = yes reflexive
… | no ¬m≡n       = no (λ where reflexive → ¬m≡n reflexive)
\end{code}

_≡?_ : (m n : ℕ) → Decide (m ≡ n)
zero   ≡? zero   = yes reflexive
zero   ≡? succ n = no (λ ())
succ m ≡? zero   = no (λ ())
succ m ≡? succ n with m ≡? n
... | yes m≡n    = yes (congruent succ m≡n)
... | no ¬m≡n    = no (λ succ-m≡succ-n → ¬m≡n (succ-injective succ-m≡succ-n))

-------------------------------------------------------------------------------

Additive monoid.

<<left-identity>>
\begin{code}
+-left-identity : ∀ {n} -> zero + n ≡ n
+-left-identity = reflexive
\end{code}

<<right-identity>>
\begin{code}
+-right-identity : ∀ {n} -> n + zero ≡ n
+-right-identity {zero}    = reflexive
+-right-identity {succ n}  = congruent succ +-right-identity
\end{code}

Renaming with explicit argument.
%
\begin{code}
+-left-unit : ∀ n -> zero + n ≡ n
+-left-unit n = reflexive

+-right-unit : ∀ n -> n + zero ≡ n
+-right-unit (zero)    = reflexive
+-right-unit (succ n)  = congruent succ (+-right-unit n)
\end{code}

<<associative>>
\begin{code}
+-associative : ∀ k {m n} -> (k + m) + n ≡ k + (m + n)
+-associative (zero)    = reflexive
+-associative (succ k)  = congruent succ (+-associative k)
\end{code}

NB. The first argument is explicit to avoid unresolved metas.

Variant with explicit arguments for rewrites.

\begin{code}
+-associative′ : ∀ k m n -> (k + m) + n ≡ k + (m + n)
+-associative′ k m n = +-associative k {m = m} {n = n}
\end{code}

TODO: |shift| is called |+-swap-succ| elsewhere.

<<commutative>>
\begin{code}
shift : ∀ m n -> m + succ n ≡ succ (m + n)
shift (zero)    n = reflexive
shift (succ m)  n = congruent succ (shift m n)

shift' : ∀ m n -> succ (m + n) ≡ m + succ n
shift' (zero)    n = reflexive
shift' (succ m)  n = congruent succ (shift' m n)

+-commutative : ∀ m n -> m + n ≡ n + m
+-commutative (zero)    n = symmetric +-right-identity
+-commutative (succ m)  n = transitive (congruent succ (+-commutative m n)) (shift' n m)
\end{code}

<<additive-monoid-reverse>>
\begin{code}
+′-left-identity : ∀ n -> zero +′ n ≡ n
+′-left-identity (zero)   = reflexive
+′-left-identity (succ n) = congruent succ (+′-left-identity n)

+′-right-identity : ∀ n -> n +′ zero ≡ n
+′-right-identity n = reflexive

+′-associative : ∀ k m n -> (k +′ m) +′ n ≡ k +′ (m +′ n)
+′-associative k m (zero)   = reflexive
+′-associative k m (succ n) = congruent succ (+′-associative k m n)
\end{code}

<<exercise-plus>>
\begin{code}
shiftᴿ : ∀ m n -> succ (m +′ n) ≡ succ m +′ n
shiftᴿ m (zero)   = reflexive
shiftᴿ m (succ n) = congruent succ (shiftᴿ m n)

+≡+′ : ∀ m n -> m + n ≡ m +′ n
+≡+′ (zero)   n = symmetric (+′-left-identity n)
+≡+′ (succ m) n = transitive (congruent succ (+≡+′ m n)) (shiftᴿ m n)
\end{code}

\begin{code}
shunt-succ : ∀ m n -> succ m + n ≡ m + succ n
shunt-succ (zero)    n = reflexive
shunt-succ (succ m)  n = congruent succ (shunt-succ m n)

shunt-succᴿ : ∀ m n -> m + succ n ≡ succ m + n
shunt-succᴿ m n = symmetric (shunt-succ m n)
\end{code}

\begin{code}
+-cancel-left : ∀ k m n → k + m ≡ k + n → m ≡ n
+-cancel-left (zero)   m n eq = eq
+-cancel-left (succ k) m n eq = +-cancel-left k m n (succ-injective eq)
\end{code}

+-cancelˡ-≡ : LeftCancellative _≡_ _+_
LeftCancellative : Op₂ A → Set _
LeftCancellative _•_ = 

\begin{code}
m+n≡0→m≡0 : ∀ {m n} → m + n ≡ 0 → m ≡ 0
m+n≡0→m≡0 {m = zero} eq = reflexive

m+n≡0→n≡0 : ∀ {m n} → m + n ≡ 0 → n ≡ 0
m+n≡0→n≡0 {m = zero} reflexive = reflexive
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Multiplicative monoid.

\begin{code}
*-left-identity : ∀ n -> succ zero * n ≡ n
*-left-identity n = +-right-identity

*-right-identity : ∀ n -> n * succ zero ≡ n
*-right-identity zero = reflexive
*-right-identity (succ n) = congruent succ (*-right-identity n)
\end{code}

\begin{code}
*-left-zero : ∀ n -> zero * n ≡ zero
*-left-zero n = reflexive

*-left-distributes-over-+ : ∀ n a b -> (a + b) * n ≡ a * n + b * n
*-left-distributes-over-+ n zero b = reflexive
*-left-distributes-over-+ n (succ a) b = transitive (congruent (λ hole → n + hole) (*-left-distributes-over-+ n a b)) (symmetric (+-associative n)) -- (a * n) (b * n)))

middle-two-interchange : ∀ a b c d -> (a + b) + (c + d) ≡ (a + c) + (b + d)
--middle-two-interchange a b c d = transitive (+-associative a b (c + d)) (transitive (congruent (λ x → a + x) (transitive (symmetric (+-associative b c d)) (transitive (congruent (λ x → x + d) (+-commutative b c)) (+-associative c b d)))) (symmetric (+-associative a c (b + d))))
middle-two-interchange a b c d = transitive (+-associative a) (transitive (congruent (λ x → a + x) (transitive (symmetric (+-associative b)) (transitive (congruent (λ x → x + d) (+-commutative b c)) (+-associative c)))) (symmetric (+-associative a))) -- a c (b + d))))

*-right-zero : ∀ n -> n * zero ≡ zero
*-right-zero (zero)   = reflexive
*-right-zero (succ n) = *-right-zero n

*-right-distributes-over-+ : ∀ n a b -> n * (a + b) ≡ n * a + n * b
*-right-distributes-over-+ (zero)   a b = reflexive
*-right-distributes-over-+ (succ n) a b = transitive (congruent (λ x → a + b + x) (*-right-distributes-over-+ n a b))
                                               (middle-two-interchange a b (n * a) (n * b))
*-associative : ∀ k m n -> (k * m) * n ≡ k * (m * n)
*-associative zero m n = reflexive
*-associative (succ k) m n =
  transitive (*-left-distributes-over-+ n m (k * m)) (congruent (λ hole → m * n + hole) (*-associative k m n))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
n*2≡n+n : ∀ n -> n *2 ≡ n + n
n*2≡n+n (zero)   = reflexive
n*2≡n+n (succ n) = congruent succ (transitive (congruent succ (n*2≡n+n n)) (shunt-succ n n))

n*succ-m≡n+n*m : {n m : ℕ} → n * (succ m) ≡ n + n * m
n*succ-m≡n+n*m {zero} {zero} = reflexive
n*succ-m≡n+n*m {zero} {succ m} = reflexive
n*succ-m≡n+n*m {succ n} {zero} = congruent succ (n*succ-m≡n+n*m {n})
n*succ-m≡n+n*m {succ n} {succ m} rewrite n*succ-m≡n+n*m {n} {succ m} rewrite n*succ-m≡n+n*m {n} {m} = congruent succ (
  succ (m + (n + (n + n * m)))
  ≡⟨ congruent succ (≡-symmetric (+-associative m)) ⟩
  succ ((m + n) + (n + n * m))
  ≡⟨ congruent (λ z → succ z + (n + n * m)) (+-commutative m n) ⟩
  succ ((n + m) + (n + n * m))
  ≡⟨ congruent succ (+-associative n) ⟩
  succ (n + (m + (n + n * m)))
  ≡⟨ ≡-symmetric (shift n (m + (n + n * m))) ⟩
  n + succ (m + (n + n * m))
  ∎)
\end{code}

\begin{code}
*-commutative : (n m : ℕ) → n * m ≡ m * n
*-commutative (zero) (zero) = reflexive
*-commutative (zero) (succ m) = *-commutative zero m
*-commutative (succ n) zero = *-commutative n zero
*-commutative (succ n) (succ m) rewrite n*succ-m≡n+n*m {n} {m} rewrite n*succ-m≡n+n*m {m} {n} = congruent succ (
  m + (n + n * m)
  ≡⟨ ≡-symmetric (+-associative m) ⟩
  (m + n) + (n * m)
  ≡⟨ congruent  (_+ (n * m)) (+-commutative m n) ⟩
  (n + m) + (n * m)
  ≡⟨ congruent ((n + m) +_) (*-commutative n m) ⟩
  (n + m) + (m * n)
  ≡⟨ +-associative n ⟩
  n + (m + m * n)
  ∎
  )
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of integer/natural division by 2.

\begin{code}
⌊/2⌋∘*2≡id : ∀ n → ⌊ (n *2) /2⌋ ≡ n
⌊/2⌋∘*2≡id (zero)   = reflexive
⌊/2⌋∘*2≡id (succ n) = congruent succ (⌊/2⌋∘*2≡id n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Monus

\begin{code}
∸-right-identity : n ∸ zero ≡ n
∸-right-identity {n = zero} = reflexive
∸-right-identity {n = succ n} = reflexive

n∸n≡zero : ∀ n -> n ∸ n ≡ zero
n∸n≡zero (zero)   = reflexive
n∸n≡zero (succ n) = n∸n≡zero n

∸-right-unit : ∀ n → n ∸ 0 ≡ n
∸-right-unit (zero)   = reflexive
∸-right-unit (succ n) = reflexive

∸-inverse : ∀ n p → (n + p) ∸ p ≡ n
∸-inverse n (zero)   = transitive (congruent₂ _∸_ (+-right-unit n) reflexive) (∸-right-unit n)
∸-inverse n (succ p) = transitive (congruent₂ _∸_ (shift n p) reflexive) (∸-inverse n p)
\end{code}

monus-sum : ∀ a b {c} → a ∸ (b + c) ≡ (a ∸ c) ∸ b
monus-sum = {!!}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Minimum. Zero.

<<min-left-zero>>
\begin{code}
⊓-left-zero : ∀ n -> zero ⊓ n ≡ zero
⊓-left-zero n = reflexive
\end{code}

<<min-right-zero>>
\begin{code}
⊓-right-zero : ∀ n -> n ⊓ zero ≡ zero
⊓-right-zero (zero)   = reflexive
⊓-right-zero (succ n) = reflexive
\end{code}

ACI.

<<min-associative>>
\begin{code}
⊓-associative : ∀ k m n -> (k ⊓ m) ⊓ n ≡ k ⊓ (m ⊓ n)
⊓-associative (zero)   m        n        = reflexive
⊓-associative (succ k) (zero)   n        = reflexive
⊓-associative (succ k) (succ m) (zero)   = reflexive
⊓-associative (succ k) (succ m) (succ n) = congruent succ (⊓-associative k m n)
\end{code}

<<min-commutative>>
\begin{code}
⊓-commutative : ∀ m n → m ⊓ n ≡ n ⊓ m
⊓-commutative (zero)   n        = symmetric (⊓-right-zero n)
⊓-commutative (succ m) (zero)   = reflexive
⊓-commutative (succ m) (succ n) = congruent succ (⊓-commutative m n)
\end{code}

<<min-idempotent>>
\begin{code}
⊓-idempotent : ∀ n → n ⊓ n ≡ n
⊓-idempotent (zero)   = reflexive
⊓-idempotent (succ n) = congruent succ (⊓-idempotent n)
\end{code}

No unit.

\begin{code}
n⊓succ-n≡n : ∀ n -> n ⊓ succ n ≡ n
n⊓succ-n≡n (zero)   = reflexive
n⊓succ-n≡n (succ n) = congruent succ (n⊓succ-n≡n n)

⊓-no-unit : ¬ ∃ (λ (u : ℕ) -> ∀ n -> u ⊓ n ≡ n)
⊓-no-unit (u , unital) with transitive (symmetric (n⊓succ-n≡n u)) (unital (succ u))
... | ()
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Maximum.

\begin{code}
either-or : ∀ m n → (m ⊔ n ≡ m) ⊎ (m ⊔ n ≡ n)
either-or (zero)   (zero)   = inl reflexive
either-or (zero)   (succ n) = inr reflexive
either-or (succ m) (zero)   = inl reflexive
either-or (succ m) (succ n) with either-or m n
... | inl m⊔n≡m = inl (congruent succ m⊔n≡m)
... | inr m⊔n≡n = inr (congruent succ m⊔n≡n)
\end{code}

Unit.

<<max-left-unit>>
\begin{code}
⊔-left-unit : ∀ n -> zero ⊔ n ≡ n
⊔-left-unit n = reflexive
\end{code}

<<max-right-unit>>
\begin{code}
⊔-right-unit : ∀ n -> n ⊔ zero ≡ n
⊔-right-unit (zero)   = reflexive
⊔-right-unit (succ n) = reflexive
\end{code}

ACI.

<<max-associative>>
\begin{code}
⊔-associative : ∀ k m n -> (k ⊔ m) ⊔ n ≡ k ⊔ (m ⊔ n)
⊔-associative (zero)   m        n        = reflexive
⊔-associative (succ k) (zero)   n        = reflexive
⊔-associative (succ k) (succ m) (zero)   = reflexive
⊔-associative (succ k) (succ m) (succ n) = congruent succ (⊔-associative k m n)
\end{code}

<<max-commutative>>
\begin{code}
⊔-commutative : ∀ m n → m ⊔ n ≡ n ⊔ m
⊔-commutative (zero)   n        = symmetric (⊔-right-unit n)
⊔-commutative (succ m) (zero)   = reflexive
⊔-commutative (succ m) (succ n) = congruent succ (⊔-commutative m n)
\end{code}

<<max-idempotent>>
\begin{code}
⊔-idempotent : ∀ n → n ⊔ n ≡ n
⊔-idempotent (zero)   = reflexive
⊔-idempotent (succ n) = congruent succ (⊔-idempotent n)
\end{code}

\begin{code}
⊔-idempotent+1 : ∀ {n} → succ n ⊔ n ≡ succ n
⊔-idempotent+1 {zero}   = reflexive
⊔-idempotent+1 {succ n} = congruent succ ⊔-idempotent+1

⊔-1+idempotent : ∀ {n} → n ⊔ succ n ≡ succ n
⊔-1+idempotent {zero}   = reflexive
⊔-1+idempotent {succ n} = congruent succ ⊔-1+idempotent
\end{code}

\begin{code}
+-distributes-over-⊔ : ∀ l m n → l + (m ⊔ n) ≡ (l + m) ⊔ (l + n)
+-distributes-over-⊔ zero m n = reflexive
+-distributes-over-⊔ (succ l) m n = congruent succ (+-distributes-over-⊔ l m n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Parity, see also VFP.Predicate.Examples.Natural.

<<even>>
\begin{code}
data even : ℕ -> Set where
  base  : even zero
  step  : {n : ℕ} → even n -> even (succ (succ n))
\end{code}

<<four-is-even>>
\begin{code}
four-is-even  :  even (succ (succ (succ (succ zero))))
four-is-even  =  step (step base)
\end{code}

<<double-is-even>>
\begin{code}
double-is-even : ∀ (n : ℕ) → even (double n)
double-is-even (zero)    = base
double-is-even (succ n)  = step (double-is-even n) 
\end{code}

<<odd>>
\begin{code}
odd : ℕ -> Set
odd n = even (succ n)
\end{code}

<<oddly>>
\begin{code}
data oddly : ℕ → Set where
  base  : oddly (succ zero)
  step  : oddly n → oddly (succ (succ n))
\end{code}

<<three-is-odd>>
\begin{code}
three-is-odd  :  odd (succ (succ (succ zero)))
three-is-odd  =  four-is-even
\end{code}

<<one-is-odd>>
\begin{code}
one-is-odd  : odd (succ zero)
one-is-odd  = step base 
\end{code}

<<there-is-an-odd-number>>
\begin{code}
there-is-odd  : Σ ℕ odd
there-is-odd  = (succ zero , step base)
\end{code}


<<even-or-odd>>
\begin{code}
data Parity (n : ℕ) : Set where
  isOdd : odd n → Parity n
  isEven : even n → Parity n
\end{code}

<<parity-type>>
\begin{code}
parity : (n : ℕ) -> even n ⊎ odd n
\end{code}
<<parity>>
\begin{code}
parity (zero)     = inl base
parity (succ n)  with parity n
parity (succ n)  | inl even-n  = inr (step even-n)
parity (succ n)  | inr odd-n   = inl odd-n
\end{code}

Or else:

<<parity-succ>>
\begin{code}
parity-succ : (n : ℕ) → even n ⊎ odd n → even (succ n) ⊎ odd (succ n)
parity-succ n (inl x) = inr (step x)
parity-succ n (inr x) = inl x
\end{code}

\begin{code}
parity' : (n : ℕ) -> even n ⊎ odd n
parity' (zero)    =  inl base
parity' (succ n)  =  let ih = parity' n in
                     parity-succ n ih
\end{code}

parity (zero)       = inl base
parity (succ zero)  = inr (step base)
parity (succ (succ n)) with parity n
... | inl even-n    = inl (step even-n)
... | inr odd-n     = inr (step odd-n)

<<even+even=even′>>
\begin{code}
even+even=even′ : (m : ℕ) -> even m -> (n : ℕ) -> even n -> even (m + n)
even+even=even′ (zero)           (base)         n even-n = even-n
even+even=even′ (succ (succ m))  (step even-m)  n even-n =
  step (even+even=even′ m even-m n even-n)
\end{code}

<<even+even=even>>
\begin{code}
even+even=even : {m n : ℕ} -> even m -> even n -> even (m + n)
even+even=even (base)         even-n = even-n
even+even=even (step even-m)  even-n =
  step (even+even=even even-m even-n)
\end{code}

\begin{code}
odd+odd : odd m -> odd n -> even (m + n)
odd+odd (step base)          odd-n = odd-n
odd+odd (step (step odd-m))  odd-n = step (odd+odd (step odd-m) odd-n)
\end{code}

<<exclusive>>
\begin{code}
exclusive : even n -> odd n -> ⊥
exclusive (step even-n) (step odd-n) = exclusive even-n odd-n
\end{code}

exclusive (base)         ()

The following is an instance of a general property: if P ⊎ Q and
furthermore P -> Q -> ⊥, then P ↔ ¬ Q and ¬ P ↔ Q.

<<even-implies-not-odd>>
\begin{code}
even→¬odd : even n -> ¬ odd n
even→¬odd = exclusive
\end{code}

<<not-odd-implies-even>>
\begin{code}
¬odd→even : ∀ n -> ¬ odd n -> even n
¬odd→even n ¬-odd-n with parity n
... | inl even-n  = even-n
... | inr odd-n   = ex-falso-quodlibet (¬-odd-n odd-n)
\end{code}

-------------------------------------------------------------------------------

\begin{code}
even-⌊/2⌋-succ : ∀ {n} -> even n -> ⌊ n /2⌋ ≡ ⌊ succ n /2⌋
even-⌊/2⌋-succ {zero} even-n = reflexive
even-⌊/2⌋-succ {succ (succ n)} (step even-n) = congruent succ (even-⌊/2⌋-succ even-n)
\end{code}

-------------------------------------------------------------------------------

\begin{code}
⌊n/2⌋+⌈n/2⌉≡n : ∀ (n : ℕ) → ⌊ n /2⌋ + ⌈ n /2⌉ ≡ n
⌊n/2⌋+⌈n/2⌉≡n zero = reflexive
⌊n/2⌋+⌈n/2⌉≡n (succ zero) = reflexive
⌊n/2⌋+⌈n/2⌉≡n (succ (succ n)) = congruent succ (
  ⌊ n /2⌋ + succ ⌊ succ n /2⌋
  ≡⟨ shift ⌊ n /2⌋ ⌈ n /2⌉ ⟩
  succ (⌊ n /2⌋ + ⌊ succ n /2⌋)
  ≡⟨ congruent succ (⌊n/2⌋+⌈n/2⌉≡n n) ⟩
  succ n
  ∎)

⌈n/2⌉+⌊n/2⌋≡n : ∀ (n : ℕ) → ⌈ n /2⌉ + ⌊ n /2⌋ ≡ n
⌈n/2⌉+⌊n/2⌋≡n n = ≡-transitive (+-commutative ⌈ n /2⌉ ⌊ n /2⌋) (⌊n/2⌋+⌈n/2⌉≡n n)
\end{code}
