Boolean values aka truth values: false and true.

Todo: move properties into submodule VFP.Boolean.Properties.

-------------------------------------------------------------------------------

\begin{code}
--{-# OPTIONS --without-K --safe --cubical-compatible #-}
{-# OPTIONS --without-K --safe #-}

module VFP.Boolean where

open import VFP.Logic
open import VFP.Equality
open import VFP.Bijection

infixr  9 _&&_
infixr  8 _||_
infixr  7 _=>_ _<=>_
\end{code}

-------------------------------------------------------------------------------

<<Boolean>>
\begin{code}
data Boolean : Set where
  false  : Boolean
  true   : Boolean
\end{code}

\begin{code}
{-# BUILTIN BOOL  Boolean  #-}
{-# BUILTIN FALSE false #-}
{-# BUILTIN TRUE  true  #-}
\end{code}

<<Boolean-mathbb>>
\begin{code}
𝔹 : Set
𝔹 = Boolean
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

<<conditional>>
\begin{code}
infix 0 if_then_else_
if_then_else_ : {A : Set} → Boolean → A → A → A
if false  then t else e = e
if true   then t else e = t
\end{code}

See also

infix 0 if_then_else_
if_then_else_ : {A : Set} → (b : Boolean) → ({So b} → A) → ({So (not b)}→ A) → A
if false  then t else e = e {yes}
if true   then t else e = t {yes}

in VFP.Searching.Binary-Search.Haskell.

The free theorem of „if_then_else_“ — the relation R is an explicit
argument to avoid unsolved meta-variables.

<<at-if>>
\begin{code}
at_if_then_else_ : ∀ {A A′ : Set} (R : A → A′ → Set)
                 → ∀ {b b′} → b ≡ b′
                 → ∀ {a a′} → R a a′
                 → ∀ {c c′} → R c c′
                 → R (if b then a else c) (if b′ then a′ else c′)
at_if_then_else_ R {b = true } reflexive q r = q
at_if_then_else_ R {b = false} reflexive q r = r
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Logical connectives.

<<not>>
\begin{code}
not : 𝔹 → 𝔹

not false = true
not true  = false
\end{code}

\begin{code}
_&&_ _||_ : 𝔹 → 𝔹 → 𝔹

false && b = false
true  && b = b

false || b = b
true  || b = true
\end{code}

<<implication>>
\begin{code}
_=>_ : 𝔹 → 𝔹 → 𝔹
false  => b = true
true   => b = b
\end{code}

\begin{code}
_<=>_ : 𝔹 → 𝔹 → 𝔹
false <=> false = true
false <=> true  = false
true  <=> false = false
true  <=> true  = true
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Interdefinability of logical connectives (using false and =>).

\begin{code}
def-not : ∀ b → b => false ≡ not b
def-not false = reflexive
def-not true  = reflexive

def-|| : ∀ b₁ b₂ → (b₂ => false) => b₁ ≡ b₁ || b₂
def-|| false false = reflexive
def-|| true  false = reflexive
def-|| false true  = reflexive
def-|| true  true  = reflexive

def-&& : ∀ b₁ b₂ → (b₂ => b₁ => false) => false ≡ b₁ && b₂
def-&& false false = reflexive
def-&& false true  = reflexive
def-&& true  false = reflexive
def-&& true  true  = reflexive
\end{code}

-------------------------------------------------------------------------------

begin{code}
private
  variable
    a b : 𝔹
end{code}

When is a Boolean expression true? This is „so true“.

<<So>>
\begin{code}
data So : Boolean → Set where
  yes : So true
\end{code}

|So b| is equivalent to |b ≡ true|.

<<from-So>>
\begin{code}
from-So : ∀ {b} → So b → b ≡ true
from-So yes = reflexive
\end{code}

\begin{code}
to-So : ∀ {b} → b ≡ true → So b
to-So reflexive = yes
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

  So false      ↔ ⊥
  So true       ↔ ⊤
  So (not a)    ↔ ¬ So a
  So (a => b)   ↔ (So a → So b)
  So (b₁ && b₂) ↔ So b₁ × So b₂
  So (b₁ || b₂) ↔ So b₁ ⊎ So b₂

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

true:

\begin{code}
true-intro : So true
true-intro = yes
\end{code}

\begin{code}
true-to : ⊤ → So true
true-to tt = yes

true-fro : So true → ⊤
true-fro yes = trivial
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

false:

\begin{code}
xfalse-elim : ∀ {b : Boolean} → So false → So b
xfalse-elim ()
\end{code}

\begin{code}
false-to : ⊥ → So false
false-to ()

false-fro : So false → ⊥
false-fro ()
\end{code}

<<false-elim>>
\begin{code}
false-elim : ∀ b → So (false => b)
false-elim b = yes
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Negation:

\begin{code}
not-intro : ∀ {b : Boolean} → So (b => false) → So (not b)
not-intro {false} yes = yes

not-elim : ∀ {b : Boolean} → So b → So (not b) → So false
not-elim yes ()
\end{code}

\begin{code}
not-to : ∀ {b : Boolean} → ¬ So b → So (not b)
not-to {false} f = yes
not-to {true}  f = ex-falso-quodlibet (f yes)

not-fro : ∀ {b : Boolean} → So (not b) → ¬ So b
not-fro {false} yes = λ ()
\end{code}

\begin{code}
not-to-fro : ∀ {b : Boolean} → ∀ (x : So (not b)) → not-to (not-fro x) ≡ x
not-to-fro {b = false} yes = reflexive

not-fro-to : ∀ {b : Boolean} → ∀ (f : ¬ So b) → ∀ so → not-fro (not-to f) so ≡ f so
not-fro-to f yes with f yes
... | ()
\end{code}

Iso needs functional extensionality.

not-iso : ∀ {b : Boolean} → So (not b) ≅ ¬ So b
not-iso = record
  { to  = not-fro
  ; fro = not-to
  ; fro∘to≡id = not-to-fro
  ; to∘fro≡id = not-fro-to }

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Implication:

<<intro>>
\begin{code}
=>-intro : ∀ {a b} → (So a → So b) → So (a => b)
=>-intro {a = false}  f = yes
=>-intro {a = true}   f = f yes
\end{code}

<<elim>>
\begin{code}
=>-elim : ∀ {a b} → So (a => b) → (So a → So b)
=>-elim yes yes = yes
\end{code}

begin{code}
xx : ∀ {b₁ b₂ : Boolean} → ∀ (so : So (b₁ => b₂)) → =>-intro {a = b₁} (=>-elim so) ≡ so
xx {false} yes = reflexive
xx {true}  yes = reflexive

yy : ∀ {b₁ b₂ : Boolean} → ∀ (f : So b₁ → So b₂) → ∀ so → =>-elim (=>-intro f) so ≡ f so
yy f yes with f yes
... | yes = reflexive
end{code}

Iso needs functional extensionality.

=>-iso : ∀ {b₁ b₂ : Boolean} → So (b₁ => b₂) ≅ (So b₁ → So b₂)
=>-iso {b₁ = b₁} = record
  { to  = =>-elim
  ; fro = =>-intro
  ; fro∘to≡id = xx {b₁ = b₁}
  ; to∘fro≡id = λ x → {!yy x!} }

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Conjunction:

\begin{code}
&&-intro : ∀ {b₁ b₂ : Boolean} → So b₁ → So b₂ → So (b₁ && b₂)
&&-intro yes yes = yes

&&-elim₁ : ∀ {b₁ b₂ : Boolean} → So (b₁ && b₂) → So b₁
&&-elim₁ {true} yes = yes

&&-elim₂ : ∀ {b₁ b₂ : Boolean} → So (b₁ && b₂) → So b₂
&&-elim₂ {true} yes = yes
\end{code}

\begin{code}
&&-to : ∀ {b₁ b₂ : Boolean} → So b₁ × So b₂ → So (b₁ && b₂)
&&-to (so-b₁ , so-b₂) = &&-intro so-b₁ so-b₂

&&-fro : ∀ {b₁ b₂ : Boolean} → So (b₁ && b₂) → So b₁ × So b₂
&&-fro so = &&-elim₁ so , &&-elim₂ so

&&-to-fro : ∀ {b₁ b₂ : Boolean} → ∀ (x : So (b₁ && b₂)) → &&-to {b₁ = b₁} (&&-fro x) ≡ x
&&-to-fro {true} {true} yes = reflexive

&&-fro-to : ∀ {b₁ b₂ : Boolean} → ∀ (x : So b₁ × So b₂) → &&-fro {b₁ = b₁} (&&-to x) ≡ x
&&-fro-to (yes , yes) = reflexive

&&-iso : ∀ {b₁ b₂ : Boolean} → So (b₁ && b₂) ≅ So b₁ × So b₂
&&-iso = record
  { to  = &&-fro
  ; fro = &&-to
  ; fro∘to≡id = &&-to-fro
  ; to∘fro≡id = &&-fro-to }
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Disjunction:

\begin{code}
||-intro₁ : ∀ {b₁ b₂ : Boolean} → So b₁ → So (b₁ || b₂)
||-intro₁ yes = yes

||-intro₂ : ∀ {b₁ b₂ : Boolean} → So b₂ → So (b₁ || b₂)
||-intro₂ {false} yes = yes
||-intro₂ {true}  yes = yes

||-elim : ∀ {b₁ b₂ b₃ : Boolean} → So (b₁ || b₂) → So (b₁ => b₃) → So (b₂ => b₃) → So b₃
||-elim {false} {.true} yes B C = C
||-elim {true} {b₂}     yes B C = B
\end{code}

\begin{code}
||-to : ∀ {b₁ b₂ : Boolean} → So b₁ ⊎ So b₂ → So (b₁ || b₂)
||-to {false} {true}  (inr yes) = yes
||-to {true}  {false} (inl yes) = yes
||-to {true}  {true}  (inl yes) = yes
||-to {true}  {true}  (inr yes) = yes

||-fro : ∀ {b₁ b₂ : Boolean} → So (b₁ || b₂) → So b₁ ⊎ So b₂
||-fro {false} {true}  yes = inr yes
||-fro {true}  {false} yes = inl yes
||-fro {true}  {true}  yes = inl yes -- OR: inr yes

||-to-fro : ∀ {b₁ b₂ : Boolean} → ∀ (x : So (b₁ || b₂)) → ||-to {b₁ = b₁} (||-fro {b₁ = b₁} x) ≡ x
||-to-fro {false} {true}  yes = reflexive
||-to-fro {true}  {false} yes = reflexive
||-to-fro {true}  {true}  yes = reflexive
\end{code}

The converse is not true. ||-to is not injective: it maps inl yes and
inr yes to yes (b₁ = true and b₂ = true).

||-fro-to : ∀ {b₁ b₂ : Boolean} → ∀ (x : So b₁ ⊎ So b₂) → ||-fro {b₁ = b₁} (||-to {b₁ = b₁} x) ≡ x
||-fro-to {false} {true}  (inr yes) = reflexive
||-fro-to {true}  {false} (inl yes) = reflexive
||-fro-to {true}  {true}  (inl yes) = reflexive
||-fro-to {true}  {true}  (inr yes) = {!!}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Excluded middle:

\begin{code}
Boolean-excluded-middle : ∀ (b : Boolean) → So (b || not b)
Boolean-excluded-middle false = yes
Boolean-excluded-middle true  = yes
\end{code}

<<double-negation>>
\begin{code}
double-negation : ∀ b → So (((b => false) => false) => b)
double-negation false  = yes
double-negation true   = yes
\end{code}

-------------------------------------------------------------------------------
