Decide a property: compute either evidence for P or for ¬ P.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Decidable where

open import VFP.Logic.Properties public
open import VFP.Natural

private
  variable
    A : Set
    P Q R : Set

infix  15 ¬?_
infixr 14 _×?_
infixr 13 _⊎?_
infixr 12 _→?_
\end{code}

-------------------------------------------------------------------------------

<<Decide>>
\begin{code}
data Decide (P : Set) : Set where
  yes  :    P -> Decide P
  no   : ¬  P -> Decide P
\end{code}

\begin{code}
map-Decide : P ↔ Q -> Decide P ↔ Decide Q
map-Decide (to & fro) =
    (λ { (yes p) → yes (to  p) ; (no ¬p)  → no (λ q → ¬p (fro q)) })
  & (λ { (yes q) → yes (fro q) ; (no ¬q)  → no (λ p → ¬q (to  p)) })
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

\begin{code}
infix 0 if_then_else_
if_then_else_ : {P A : Set} -> Decide P → ({P} -> A) → ({¬ P} -> A) → A
if yes p then t else e = t { p}
if no ¬p then t else e = e {¬p}
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

A predicate P is decidable if P a is decidable for all a.

\begin{code}
Decidable : (A -> Set) -> Set
Decidable P = ∀ a -> Decide (P a)
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Iverson bracket (see https://en.wikipedia.org/wiki/Iverson_bracket).

\begin{code}
[_] : Decide P → ℕ
[ yes p ] = 1
[ no ¬p ] = 0

#_ : Decide P → ℕ
# yes p = 1
# no ¬p = 0
\end{code}

-------------------------------------------------------------------------------

Logical connectives.

\begin{code}
⊥? : Decide ⊥
⊥? = no (λ bot → bot)
\end{code}

\begin{code}
⊤? : Decide ⊤
⊤? = yes trivial
\end{code}

<<negation>>
\begin{code}
¬?_ : Decide P -> Decide (¬ P)
¬? yes p  = no (λ ¬p → ¬p p)
¬? no ¬p  = yes ¬p
\end{code}

<<conjunction>>
\begin{code}
_×?_ : Decide P -> Decide Q -> Decide (P × Q)
yes p  ×? yes q  = yes (p , q)
yes p  ×? no ¬q  = no (contrapositive outr ¬q)
no ¬p  ×? yes q  = no (contrapositive outl ¬p)
no ¬p  ×? no ¬q  = no (contrapositive outl ¬p)  -- or: |contrapositive outr ¬q|
\end{code}

<<disjunction>>
\begin{code}
_⊎?_ : Decide P -> Decide Q -> Decide (P ⊎ Q)
yes p  ⊎? yes q  = yes (inl p)  -- or: |yes (inr q)|
yes p  ⊎? no ¬q  = yes (inl p)
no ¬p  ⊎? yes q  = yes (inr q)
no ¬p  ⊎? no ¬q  = no (case ¬p ¬q)
\end{code}

<<implication>>
\begin{code}
_→?_ : Decide P -> Decide Q -> Decide (P → Q)
yes p  →? yes q  = yes (λ _ → q)
yes p  →? no ¬q  = no (λ f → ¬q (f p))
no ¬p  →? yes q  = yes (λ _ → q)
no ¬p  →? no ¬q  = yes (λ p → ex-falso-quodlibet (¬p p))
\end{code}

third eqn  -- or: term below

_     →? yes q = yes …
no ¬p →? _     = yes …
yes p →? no ¬q = no  …

begin{code}
infixr  8 _∨_
infixr  9 _∧_
infixr 10 ~_

_∨_ : Decide P -> Decide Q -> Decide (P ⊎ Q)
yes p ∨ yes q = yes (inl p)  -- or: |yes (inr q)|
yes p ∨ no ¬q = yes (inl p)
no ¬p ∨ yes q = yes (inr q)
no ¬p ∨ no ¬q = no (λ { (inl p) → ¬p p ; (inr q) → ¬q q })

_∧_ : Decide P -> Decide Q -> Decide (P × Q)
yes p ∧ yes q = yes (p , q)
yes p ∧ no ¬q = no (λ { (p , q) → ¬q q })
no ¬p ∧ yes q = no (λ { (p , q) → ¬p p })
no ¬p ∧ no ¬q = no (λ { (p , q) → ¬q q })

~_ : Decide P -> Decide (¬ P)
~ yes p = no (λ ¬p → ¬p p)
~ no ¬p = yes ¬p
end{code}

Todo: same as
  map-Decide : P ↔ Q -> Decide P ↔ Decide Q

begin{code}
decide-lemma : ∀ {A B : Set} -> A ↔ B -> Decide A ↔ Decide B
decide-lemma (to & fro) = (λ where
                             (yes x) → yes (to x)
                             (no x)  → no (λ z → x (fro z)))
                        & (λ where
                              (yes x) → yes (fro x)
                              (no x)  → no (λ z → x (to z)))
end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Todo: move to Properties.

<<tertium-non-datur>>
\begin{code}
tertium-non-datur : Decide P -> P ⊎ ¬ P
tertium-non-datur (yes p)  = inl p
tertium-non-datur (no ¬p)  = inr ¬p
\end{code}

<<double-negation>>
\begin{code}
double-negation : Decide P → (¬ ¬ P → P)
double-negation (yes p)  = λ ¬¬p → p
double-negation (no ¬p)  = λ ¬¬p → ex-falso-quodlibet (¬¬p ¬p)
\end{code}

De Morgan:

<<De-Morgan>>
\begin{code}
negate-× : Decide P → Decide Q → ¬ (P × Q) ↔ ¬ P ⊎ ¬ Q
negate-× (yes p)  (yes q)  = (λ z → inl (λ x → z (x , q))) & (λ { (inl x) x₁ → x (outl x₁) ; (inr x) x₁ → x (outr x₁) })
negate-× (yes p)  (no ¬q)  = (λ z → inr (λ x → z (p , x))) & (λ _ x → ¬q (outr x))
negate-× (no ¬p)  (yes q)  = (λ z → inl (λ x → z (x , q))) & (λ _ x → ¬p (outl x))
negate-× (no ¬p)  (no ¬q)  = (λ _ → inl ¬p) & (λ _ x → ¬q (outr x))
\end{code}

-------------------------------------------------------------------------------
