Properties of propositional and Leibniz equality.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Equality.Properties where

open import VFP.Equality public
open import VFP.Equivalence
open import VFP.Function
open import VFP.Logic.Properties

private
  variable
    A B C D E : Set
\end{code}

-------------------------------------------------------------------------------

Properties of propositional equality.

<<properties>>
\begin{code}
≡-symmetric : {a b : A} -> a ≡ b -> b ≡ a
-- ≡-symmetric a≡b = {!!}
≡-symmetric ≡-reflexive = ≡-reflexive

≡-transitive : {a b c : A} -> a ≡ b -> b ≡ c -> a ≡ c
-- ≡-transitive a≡b b≡c = {!!}
≡-transitive ≡-reflexive ≡-reflexive = ≡-reflexive

≡-congruent : (f : A -> B) -> {a b : A} -> a ≡ b -> f a ≡ f b
-- ≡-congruent f a≡b = {!!}
≡-congruent f ≡-reflexive = ≡-reflexive
\end{code}

\begin{code}
≡-congruent₂ : (g : A -> B -> C) -> {a₁ a₂ : A} -> a₁ ≡ a₂ -> {b₁ b₂ : B} -> b₁ ≡ b₂ -> g a₁ b₁ ≡ g a₂ b₂
≡-congruent₂ g ≡-reflexive ≡-reflexive = ≡-reflexive

≡-congruent₃ : (g : A -> B -> C -> D) -> {a₁ a₂ : A} -> a₁ ≡ a₂ -> {b₁ b₂ : B} -> b₁ ≡ b₂ -> {c₁ c₂ : C} -> c₁ ≡ c₂ -> g a₁ b₁ c₁ ≡ g a₂ b₂ c₂
≡-congruent₃ g ≡-reflexive ≡-reflexive ≡-reflexive = ≡-reflexive

≡-congruent₄ : (g : A -> B -> C -> D -> E) -> {a₁ a₂ : A} -> a₁ ≡ a₂ -> {b₁ b₂ : B} -> b₁ ≡ b₂ -> {c₁ c₂ : C} -> c₁ ≡ c₂ -> {d₁ d₂ : D} -> d₁ ≡ d₂ -> g a₁ b₁ c₁ d₁ ≡ g a₂ b₂ c₂ d₂
≡-congruent₄ g ≡-reflexive ≡-reflexive ≡-reflexive ≡-reflexive = ≡-reflexive
\end{code}

Substitutivity: any two related elements satisfy exactly the same set
of unary relations.

The defining property of Leibniz equality:

\begin{code}
≡-substitutive : ∀ (P : A → Set) → {a b : A} -> a ≡ b → P a → P b
≡-substitutive P reflexive p = p

-- this seems to be more useful!?
≡-substitutive˘ : ∀ (P : A → Set) → {a b : A} -> a ≡ b → P b → P a
≡-substitutive˘ P reflexive p = p

≡-substitutive′ : {a b : A} -> a ≡ b → ∀ (P : A → Set) → P a → P b
≡-substitutive′ reflexive P p = p

≡-substitutive₂ : ∀ (P : A -> B -> Set) -> {a₁ a₂ : A} -> a₁ ≡ a₂ -> {b₁ b₂ : B} -> b₁ ≡ b₂ -> P a₁ b₁ -> P a₂ b₂
≡-substitutive₂ P ≡-reflexive ≡-reflexive p = p
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Renamings.

\begin{code}
symmetric = ≡-symmetric
transitive = ≡-transitive
congruent = ≡-congruent
congruent₂ = ≡-congruent₂
congruent₃ = ≡-congruent₃
congruent₄ = ≡-congruent₄
substitutive = ≡-substitutive
substitutive˘ = ≡-substitutive˘
substitutive₂ = ≡-substitutive₂
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
≡-equivalence : is-Equivalence {A = A} _≡_
≡-equivalence = record
  { reflexive  = ≡-reflexive
  ; symmetric  = ≡-symmetric
  ; transitive = ≡-transitive }
\end{code}

-------------------------------------------------------------------------------

Properties of inequality.

\begin{code}
≢-irreflexive : {a : A} → ¬ a ≢ a
≢-irreflexive a≢a = a≢a reflexive

≢-symmetric : {a b : A} → a ≢ b → b ≢ a
≢-symmetric = contrapositive symmetric
\end{code}

As an aside, an *apartness relation* is a symmetric, irreflexive
binary relation with the additional condition that if two elements are
apart, then any other element is apart from at least one of them
(co-transitivity).

#-cotransitive : {a b c : A} -> a # c → a # b ⊎ b # c

The complement of an apartness relation is an equivalence relation.
    
-------------------------------------------------------------------------------

Properties of Leibniz equality.

<<Leibniz-properties>>
\begin{code}
≐-reflexive : {a : A} -> a ≐ a
≐-reflexive = λ P -> identity

≐-symmetric : {a b : A} -> a ≐ b -> b ≐ a
≐-symmetric {a = a} a≐b = λ P → a≐b (λ x -> P x -> P a) identity

≐-transitive : {a b c : A} -> a ≐ b -> b ≐ c -> a ≐ c
≐-transitive a≐b b≐c = λ P → b≐c P ∘ a≐b P

≐-congruent : {a b : A} -> (f : A -> B) -> a ≐ b -> f a ≐ f b
≐-congruent f a≐b = λ P → a≐b (λ x → P (f x))
\end{code}

-------------------------------------------------------------------------------

Equality as the smallest reflexive relation: R reflexive implies _≑_ ⊆ R.

<<smallest-reflexive>>
\begin{code}
smallest-reflexive : ∀ (R : A → A → Set) → (∀ {x} → R x x) → {a b : A} -> a ≑ b -> R a b
smallest-reflexive R R-reflexive a≑b = a≑b R R-reflexive
\end{code}

\begin{code}
≑-reflexive : {a : A} -> a ≑ a
≑-reflexive = λ R r → r

≑-symmetric : {a b : A} -> a ≑ b -> b ≑ a
≑-symmetric a≐b = λ R → a≐b (λ a b → R b a)

≑-transitive : {a b c : A} -> a ≑ b -> b ≑ c -> a ≑ c
≑-transitive {c = c} a≐b b≐c = λ R r → a≐b (λ a b → R b c -> R a c) (λ R-x-c → R-x-c) (b≐c R r)
\end{code}

-------------------------------------------------------------------------------

Propositional equality is the same as Leibniz equality.

<<equivalent>>
\begin{code}
≡-implies-≐ : ∀ {a b : A} → a ≡ b → a ≐ b
≡-implies-≐ ≡-reflexive = ≐-reflexive

≐-implies-≡ : ∀ {a b : A} → a ≐ b → a ≡ b
≐-implies-≡ {a = a} a≐b = a≐b (λ x -> a ≡ x) ≡-reflexive
\end{code}

\begin{code}
≡-implies-≑ : ∀ {a b : A} → a ≡ b → a ≑ b  -- easy, shows that ≡ is the smallest reflexive relation
≡-implies-≑ reflexive = ≑-reflexive

≑-implies-≡ : ∀ {a b : A} → a ≑ b → a ≡ b  -- easy as ≑ is in the intersection of all reflexive relations
≑-implies-≡ {a = a} a≐b = a≐b _≡_ reflexive
\end{code}
