\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.List.Properties where

open import VFP.Natural -- public
open import VFP.List public
open import VFP.Equality.Reasoning
open import VFP.Natural.Properties

private
    variable
        A B : Set
\end{code}

-------------------------------------------------------------------------------

Free monoid.

<<free-monoid>>
\begin{code}
⧺-left-unit : ∀ (xs : List A) → [] ⧺ xs ≡ xs
⧺-left-unit xs = reflexive

⧺-right-unit : ∀ (xs : List A) → xs ⧺ [] ≡ xs
⧺-right-unit []           = reflexive
⧺-right-unit (x ∷ xs)     = congruent (x ∷_) (⧺-right-unit xs) 

⧺-associative : ∀ (xs ys zs : List A) → (xs ⧺ ys) ⧺ zs ≡ xs ⧺ (ys ⧺ zs)
⧺-associative []       ys zs     = reflexive
⧺-associative (x ∷ xs) ys zs     = congruent (x ∷_) (⧺-associative xs ys zs)
\end{code}

Some convenience definitions.

\begin{code}
⧺-associative° : ∀ (xs : List A) {ys zs : List A} → xs ⧺ (ys ⧺ zs) ≡ (xs ⧺ ys) ⧺ zs
⧺-associative° xs = symmetric (⧺-associative xs _ _)

⧺-right-unit° : ∀ {xs : List A} → xs ≡ xs ⧺ []
⧺-right-unit° = symmetric (⧺-right-unit _)
\end{code}

⧺ is a monoid homomorphism.

\begin{code}
map-homo-[] : {A B : Set} → (f : A -> B) -> map f [] ≡ []
map-homo-[] f = reflexive

map-homo-⧺ : {A B : Set} → (f : A -> B) -> ∀ (xs ys : List A) → map f (xs ⧺ ys) ≡ map f xs ⧺ map f ys
map-homo-⧺ f []       ys = reflexive
map-homo-⧺ f (x ∷ xs) ys = congruent (λ ◌ → f x ∷ ◌) (map-homo-⧺ f xs ys)
\end{code}

-------------------------------------------------------------------------------

Correctness of |fast-reverse|.


broken : ∀ (xs : List A) → fast-reverse xs ≡ slow-reverse xs
broken []       = reflexive
broken (x ∷ xs) = let ih = broken xs in
                  {!!}


\begin{code}
xsnoc-lemma : ∀ (xs ys : List A) (y : A) →
  xs ⧺ y ∷ ys ≡ xsnoc xs y ⧺ ys
xsnoc-lemma [] ys y = reflexive
xsnoc-lemma (x ∷ xs) ys y =
  proof
      (x ∷ xs) ⧺ (y ∷ ys)
    ≡⟨ by-definition ⟩
      x ∷ (xs ⧺ (y ∷ ys))
    ≡⟨ congruent (x ∷_) (xsnoc-lemma xs ys y) ⟩    
      x ∷ xsnoc xs y ⧺ ys
    ∎
\end{code}

<<fast-reverse-lemma-type>>
\begin{code}
lemma : ∀ (xs ys : List A) →
  reverse-append xs ys ≡ slow-reverse xs ⧺ ys
\end{code}
<<fast-reverse-lemma-proof>>
\begin{code}
lemma [] ys =
  proof
    reverse-append [] ys
  ≡⟨ by-definition ⟩
    ys
  ≡⟨ ⧺-left-unit ys ⟩          -- |[]| is the left unit of |⧺|
    [] ⧺ ys
  ≡⟨ by-definition ⟩                     
    slow-reverse [] ⧺ ys
  ∎
lemma (x ∷ xs) ys =
  proof
     reverse-append (x ∷ xs) ys
   ≡⟨ by-definition ⟩                    
     reverse-append xs (x ∷ ys)
   ≡⟨ lemma xs (x ∷ ys) ⟩
       -- induction hypothesis
     slow-reverse xs ⧺ (x ∷ ys)
   ≡⟨ by-definition ⟩                     
     slow-reverse xs ⧺ ((x ∷ []) ⧺ ys)
   ≡⟨ symmetric (⧺-associative (slow-reverse xs) (x ∷ []) ys) ⟩
       -- |⧺| is associative
     (slow-reverse xs ⧺ (x ∷ [])) ⧺ ys    
   ≡⟨ by-definition ⟩                     
     slow-reverse (x ∷ xs) ⧺ ys
  ∎
\end{code}

<<fast-reverse-theorem>>
\begin{code}
correct : ∀ (xs : List A) →
  fast-reverse xs ≡ slow-reverse xs
correct xs =
  proof
    fast-reverse xs
  ≡⟨ by-definition ⟩                     -- holds by definition
    reverse-append xs []
  ≡⟨ lemma xs [] ⟩                       -- apply lemma
    slow-reverse xs ⧺ []
  ≡⟨ ⧺-right-unit (slow-reverse xs) ⟩    -- |[]| is the right unit of |⧺|
    slow-reverse xs
  ∎
\end{code}

-------------------------------------------------------------------------------

\begin{code}
xsnoc-append-lemma : (xs ys : List A) (y : A) →  xsnoc (xs ⧺ ys) y ≡ xs ⧺ xsnoc ys y
xsnoc-append-lemma [] ys x = reflexive
xsnoc-append-lemma (x ∷ xs) ys y =
  proof
    xsnoc ((x ∷ xs) ⧺ ys) y
  ≡⟨ by-definition ⟩    
    x ∷ xsnoc (xs ⧺ ys) y
  ≡⟨ congruent (x ∷_) (xsnoc-append-lemma xs ys y) ⟩
    x ∷ xs ⧺ xsnoc ys y
  ∎  

\end{code}

Properties of |reverse|: |reverse| is a monoid homomorphism, it
preserves the binary operation (•) and its identity element (e).

\begin{code}
reverse-homo-e : {A : Set} → reverse {A} [] ≡ []
reverse-homo-e = reflexive

reverse-homo-• : ∀ (xs ys : List A) → reverse (xs ⧺ ys) ≡ reverse ys ⧺ reverse xs
reverse-homo-• [] ys =
  proof
    reverse ([] ⧺ ys)
  ≡⟨ symmetric (⧺-right-unit (reverse ys)) ⟩
    reverse ys ⧺ []
  ≡⟨ by-definition ⟩
    reverse ys ⧺ reverse []
  ∎
reverse-homo-• (x ∷ xs) ys =
  proof
    reverse ((x ∷ xs) ⧺ ys)
  ≡⟨ by-definition ⟩
    (reverse (xs ⧺ ys)) ⧺ [ x ]
  ≡⟨ congruent (_⧺ _) (reverse-homo-• xs ys) ⟩
    (reverse ys ⧺ reverse xs) ⧺ (x ∷ [])
  ≡⟨ ⧺-associative (reverse ys) (reverse xs) (x ∷ []) ⟩
    (reverse ys ⧺ reverse (x ∷ xs))
  ∎
\end{code}

|reverse| is involutive.

\begin{code}
reverse-involutive : ∀ (xs : List A) → reverse (reverse xs) ≡ xs
reverse-involutive [] = reflexive
reverse-involutive (x ∷ xs) =
  proof
    reverse (reverse (x ∷ xs))
  ≡⟨ reverse-homo-• (reverse xs) (x ∷ []) ⟩
    x ∷ reverse (reverse xs)
  ≡⟨ congruent (_ ∷_) (reverse-involutive xs) ⟩
    x ∷ xs
  ∎
\end{code}

-------------------------------------------------------------------------------

Properties of |length|: |reverse| is a monoid homomorphism, it
preserves the binary operation (•) and its identity element (e).

<<length-homo>>
\begin{code}
--length-homo-e : length [] ≡ 0
length-homo-e : {A : Set} → length {Elem = A} [] ≡ 0
length-homo-e = reflexive

length-homo-• : ∀ (xs ys : List A) → length (xs ⧺ ys) ≡ length xs + length ys
length-homo-• []       ys = reflexive
length-homo-• (x ∷ xs) ys = congruent succ (length-homo-• xs ys)
\end{code}

OR: length-preserves-unit, length-preserves-binop.

|length| and |reverse|.

<<length-reverse>>
\begin{code}
length-xsnoc : ∀ (xs : List A) (y : A)  → length (xsnoc xs y) ≡ 1 + length xs
length-xsnoc [] y = reflexive
length-xsnoc (x ∷ xs) y = congruent succ (length-xsnoc xs y)

length-reverse : ∀ (xs : List A) → length (reverse xs) ≡ length xs
length-reverse []       = reflexive
length-reverse (x ∷ xs) =
  proof
    length (reverse (x ∷ xs))
  ≡⟨ by-definition ⟩
    length ((reverse xs) ⧺ [ x ])
  ≡⟨ length-homo-• (reverse xs) (x ∷ []) ⟩
    length (reverse xs) + length (x ∷ [])
  ≡⟨ by-definition ⟩
    length (reverse xs) + 1
  ≡⟨ congruent (_+ 1) (length-reverse xs) ⟩
    length xs + 1
  ≡⟨ +-commutative (length xs) 1 ⟩
    1 + length xs
  ≡⟨ by-definition ⟩ 
    length (x ∷ xs)
  ∎
\end{code}

\begin{code}
length-map : ∀ (f : A → B) → ∀ (xs : List A) → length (map f xs) ≡ length xs
length-map f []       = reflexive
length-map f (x ∷ xs) = congruent succ (length-map f xs)
\end{code}

-------------------------------------------------------------------------------

Properties of (for)all.

\begin{code}
all-map : (P : B -> Set) -> (f : A -> B) ->
  ∀ {xs} -> all (λ x -> P (f x)) xs -> all P (map f xs)
all-map P f []       = []
all-map P f (p ∷ ps) = p ∷ all-map P f ps

all-map° : (P : B -> Set) -> (f : A -> B) ->
  ∀ {xs} -> all P (map f xs) -> all (λ x -> P (f x)) xs
all-map° P f {xs = []}     []       = []
all-map° P f {xs = x ∷ xs} (p ∷ ps) = p ∷ all-map° P f ps

all-monotone : {P Q : B -> Set} -> (∀ {b} -> P b -> Q b) ->
  ∀ {xs} -> all P xs -> all Q xs
all-monotone P→Q [] = []
all-monotone P→Q (x ∷ xxx) = P→Q x ∷ all-monotone P→Q xxx

all-⧺ : {P : A -> Set} -> ∀ {xs ys} -> all P xs -> all P ys -> all P (xs ⧺ ys)
all-⧺ []       qs = qs
all-⧺ (p ∷ ps) qs = p ∷ all-⧺ ps qs
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of exist.

\begin{code}
exists-map : (P : B -> Set) -> (f : A -> B) ->
  ∀ {xs} -> exists (λ x -> P (f x)) xs -> exists P (map f xs)
exists-map P f (head x)   = head x
exists-map P f (tail pos) = tail (exists-map P f pos)

exists-map° : (P : B -> Set) -> (f : A -> B) ->
  ∀ {xs} -> exists P (map f xs) -> exists (λ x -> P (f x)) xs
exists-map° P f {xs = x ∷ xs} (head x₁) = head x₁
exists-map° P f {xs = x ∷ xs} (tail xx) = tail (exists-map° P f xx)

exists-⧺₁ : (P : A -> Set) -> ∀ {xs ys} -> exists P xs -> exists P (xs ⧺ ys)
exists-⧺₁ P (head x) = head x
exists-⧺₁ P (tail p) = tail (exists-⧺₁ P p)

exists-⧺₂ : (P : A -> Set) -> ∀ {xs ys} -> exists P ys -> exists P (xs ⧺ ys)
exists-⧺₂ P {xs = []}     p = p
exists-⧺₂ P {xs = x ∷ xs} p = tail (exists-⧺₂ P p)

open import VFP.Decidable
exists-⧺ : (P : A -> Set) -> ∀ {xs ys} -> exists P (xs ⧺ ys) -> exists P xs ⊎ exists P ys
exists-⧺ P {xs = []} p = inr p
exists-⧺ P {xs = x ∷ xs} (head x₁) = inl (head x₁)
exists-⧺ P {xs = x ∷ xs} (tail p) with exists-⧺ P {xs = xs} p
... | inl x₁ = inl (tail x₁)
... | inr y = inr y
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Properties of _∈′_.

\begin{code}
∈-⧺₁ : ∀ {x : A} {xs ys} -> x ∈′ xs -> x ∈′ xs ⧺ ys  -- rename: ∈-⧺-inl, ∈-⧺-inr, ∈-⧺-case
∈-⧺₁ = exists-⧺₁ _

∈-⧺₂ : ∀ {x : A} {xs ys} -> x ∈′ ys -> x ∈′ xs ⧺ ys
∈-⧺₂ = exists-⧺₂ _

∈-⧺ : ∀ {x : A} {xs ys} -> x ∈′ xs ⧺ ys -> x ∈′ xs ⊎ x ∈′ ys
∈-⧺ = exists-⧺ _
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Relating exists and (for)all.

\begin{code}
decide-all : (P : A -> Set) -> (∀ x -> Decide (P x)) -> ∀ xs -> Decide (all P xs)
decide-all P decide [] = yes []
decide-all P decide (x ∷ xs) with decide x | decide-all P decide xs
... | yes x₁ | yes x₂ = yes (x₁ ∷ x₂)
... | yes x₁ | no x₂ = no (λ { (x ∷ x₃) → x₂ x₃ })
... | no x₁ | yes x₂ = no (λ { (x ∷ x₃) → x₁ x })
... | no x₁ | no x₂ = no (λ { (x ∷ x₃) → x₂ x₃ })

decide-exists : (P : A -> Set) -> (∀ x -> Decide (P x)) -> ∀ xs -> Decide (exists P xs)
decide-exists P decide [] = no (λ ())
decide-exists P decide (x ∷ xs) with decide x | decide-exists P decide xs
... | yes x₁ | yes x₂ = yes (head x₁)
... | yes x₁ | no x₂ = yes (head x₁)
... | no x₁ | yes x₂ = yes (tail x₂)
... | no x₁ | no x₂ = no (λ { (head x) → x₁ x ; (tail zz) → x₂ zz })

De-Morgan : (P : A -> Set) -> ∀ xs -> ¬ exists P xs -> all (λ x -> ¬ P x) xs
De-Morgan P [] f = []
De-Morgan P (x ∷ xs) f = (λ x → f (head x)) ∷ De-Morgan P xs (λ z → f (tail z))

-- ¬ all P xs -> exists (λ x -> ¬ P x) xs
\end{code}

\begin{code}
eithers : (P Q : A -> Set) -> (∀ x -> P x ⊎ Q x) -> (∀ xs -> exists P xs ⊎ all Q xs)
eithers P Q either [] = inr []
eithers P Q either (x ∷ xs) with either x | eithers P Q either xs
... | inl x₁ | inl x₂ = inl (head x₁)
... | inl x₁ | inr y = inl (head x₁)
... | inr y | inl x₁ = inl (tail x₁)
... | inr y | inr y₁ = inr (y ∷ y₁)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
private
    variable
        P Q : A -> Set

filter-filter : (P? : Decidable P) -> (Q? : Decidable Q) -> ∀ xs -> filter Q? (filter P? xs) ≡ filter (λ a -> P? a ×? Q? a) xs
filter-filter P? Q? [] = reflexive
filter-filter P? Q? (x ∷ xs) with P? x
filter-filter P? Q? (x ∷ xs) | yes p with Q? x
... | yes q = congruent (x ∷_) (filter-filter P? Q? xs)
... | no ¬q = filter-filter P? Q? xs
filter-filter P? Q? (x ∷ xs) | no ¬p with Q? x
... | yes q = filter-filter P? Q? xs
... | no ¬q = filter-filter P? Q? xs
\end{code}

\begin{code}
all-filter : (P? : Decidable P) → ∀ xs -> all P (filter P? xs)
all-filter P? [] = []
all-filter P? (x ∷ xs) with P? x
... | yes p      = p ∷ all-filter P? xs
... | no ¬p      =     all-filter P? xs
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Canned induction aka free theorems (for predicates and relations).

<<foldr-induction>>
\begin{code}
foldr-induction : {A : Set}
                → {B : Set} (P : B → Set)
                → {_◁_ : A -> B -> B} → (∀ (x : A) → ∀ {r} → P r → P (x ◁ r))
                → {b : B} → P b
                → (xs : List A) → P (foldr _◁_ b xs)
foldr-induction P Step Base []        = Base
foldr-induction P Step Base (x ∷ xs)  = Step x (foldr-induction P Step Base xs)
\end{code}

<<foldr-induction-2>>
\begin{code}
foldr-induction₂  : {A : Set}
                  → {B₁ B₂ : Set} → (R : B₁ → B₂ → Set)
                  → {_◁₁_ : A -> B₁ -> B₁} {_◁₂_ : A -> B₂ -> B₂} → (∀ (x : A) → ∀ {r₁ r₂} → R r₁ r₂ → R (x ◁₁ r₁) (x ◁₂ r₂))
                  → {b₁ : B₁} {b₂ : B₂} → R b₁ b₂
                  → (xs : List A) → R (foldr _◁₁_ b₁ xs) (foldr _◁₂_ b₂ xs)
foldr-induction₂ R Step Base []        = Base
foldr-induction₂ R Step Base (x ∷ xs)  = Step x (foldr-induction₂ R Step Base xs)
\end{code}

See also, VFP.Experimental.Theorems-for-free: foldr-induction₂ is a
simplified variant of [foldr] : [Foldr] foldr foldr, fixing the type
parameter A.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
foldl-induction : {A B : Set} {_◁_ : B → A → B} {s : B} → (P : B → Set) →
    (∀ (x : A) → ∀ {r} → P r → P (r ◁ x)) → P s → (xs : List A) → P (foldl _◁_ s xs)
foldl-induction P Step Base [] = Base
foldl-induction P Step Base (x ∷ xs) = foldl-induction P Step (Step x Base) xs
\end{code}

\begin{code}
foldl-induction₂ : {A B₁ B₂ : Set}
                → (R : B₁ → B₂ → Set)
                → {_◁₁_ : B₁ → A → B₁} {_◁₂_ : B₂ → A → B₂} → (∀ (x : A) → ∀ {r₁ r₂} → R r₁ r₂ → R (r₁ ◁₁ x) (r₂ ◁₂ x))
                → {b₁ : B₁} {b₂ : B₂} → R b₁ b₂ → (xs : List A) → R (foldl _◁₁_ b₁ xs) (foldl _◁₂_ b₂ xs)
foldl-induction₂ R Step Base [] = Base
foldl-induction₂ R Step Base (x ∷ xs) = foldl-induction₂ R Step (Step x Base) xs
\end{code}

-------------------------------------------------------------------------------
