Natural numbers: unary representation aka Peano numerals.

-------------------------------------------------------------------------------

\begin{code}
--{-# OPTIONS --without-K --safe --cubical-compatible #-}
{-# OPTIONS --without-K --safe #-}

module VFP.Natural where

open import VFP.Boolean

infixl 8 _+_ _+′_ _∸_
infixl 9 _*_ _*2
infixr 7 _⊓_ _⊔_
infix  6 _≤_ _≥_ _<_ _>_
\end{code}

-------------------------------------------------------------------------------

<<Natural>>
\begin{code}
data Natural : Set where
  zero  :  Natural
  succ  :  Natural → Natural
\end{code}

\begin{code}
{-# BUILTIN NATURAL Natural #-}
\end{code}

The |BUILTIN| pragma enables us to use |0|, |1| etc as naturals.

ℕ : Set

<<blackboard>>
\begin{code}
ℕ = Natural
\end{code}

<<five>>
\begin{code}
five  :  Natural
five  =  succ (succ (succ (succ (succ zero))))
\end{code}

<<seven>>
\begin{code}
seven  :  Natural
seven  =  succ (succ (succ (succ (succ (succ (succ zero))))))
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Folding natural numbers.

<<recurse>>
\begin{code}
recurse : ∀ {A : Set} -> (A -> A) -> A -> ℕ -> A
recurse s z (zero)    = z
recurse s z (succ n)  = s (recurse s z n)
\end{code}

<<iterate>>
\begin{code}
iterate : ∀ {A : Set} -> (A -> A) -> A -> (ℕ -> A)
iterate s z (zero)    = z
iterate s z (succ n)  = iterate s (s z) n
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Basic arithmetic.

\begin{code}
pred : ℕ -> ℕ
pred (zero)   = zero
pred (succ n) = n
\end{code}

<<plus>>
\begin{code}
_+_ : Natural -> Natural -> Natural
zero    + n = n
succ m  + n = succ (m + n)
\end{code}

<<plus-reverse>>
\begin{code}
_+′_ : Natural -> Natural -> Natural
m +′ zero    = m
m +′ succ n  = succ (m +′ n)
\end{code}

\begin{code}
_+″_ : Natural -> Natural -> Natural
zero   +″ zero   = zero
zero   +″ succ n = succ n
succ m +″ zero   = succ m
succ m +″ succ n = succ (succ (m +″ n))
\end{code}

Monus.

<<monus>>
\begin{code}
_∸_ : Natural -> Natural -> Natural
zero    ∸ n       = zero
succ m  ∸ zero    = succ m
succ m  ∸ succ n  = m ∸ n
\end{code}

<<times>>
\begin{code}
_*_ : Natural -> Natural -> Natural
zero    * n = zero
succ m  * n = n + (m * n)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Multiplication and division by 2 aka doubling and halving.

<<times2>>
\begin{code}
_*2  : Natural -> Natural
zero   *2 = zero
succ n *2 = succ (succ (n *2))
\end{code}

<<double>>
\begin{code}
double : Natural → Natural
double (zero)    = zero
double (succ n)  = succ (succ (double n))
\end{code}

<<halve>>
\begin{code}
halve : ℕ -> ℕ
halve (zero)          = zero
halve (succ zero)     = zero
halve (succ (succ n)) = succ (halve n)
\end{code}

Division by 2, rounded downwards and upwards.

\begin{code}
⌊_/2⌋ : ℕ → ℕ
⌊ 0 /2⌋             = 0
⌊ 1 /2⌋             = 0
⌊ succ (succ n) /2⌋ = succ ⌊ n /2⌋

⌈_/2⌉ : ℕ → ℕ
⌈ n /2⌉ = ⌊ succ n /2⌋
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Naïve exponentiation.

<<exponentiation>>
\begin{code}
_^_ : Natural → Natural → Natural
x ^  zero    = succ zero
x ^  succ n  = x * (x ^ n)
\end{code}

\begin{code}
2^_ : Natural → Natural
2^ zero   = 1
2^ succ n = (2^ n) *2
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Comparison.

<<ordering>>
\begin{code}
_≤_ : Natural -> Natural -> Boolean

zero    ≤ n       =  true
succ m  ≤ zero    =  false
succ m  ≤ succ n  =  m ≤ n
\end{code}

<<reverse-ordering>>
\begin{code}
_≥_ : Natural -> Natural -> Boolean
m ≥ n = n ≤ m
\end{code}

m < n = not (m ≥ n)

<<strict-ordering>>
\begin{code}
_<_ : Natural -> Natural -> Boolean
m       < zero    = false
zero    < succ n  = true
succ m  < succ n  = m < n
\end{code}

\begin{code}
_>_ : Natural -> Natural -> Boolean
m > n = not (m ≤ n)
\end{code}

Minimum and maximum (Order defines ↑).

<<minimum>>
\begin{code}
_⊓_ : ℕ → ℕ → ℕ
zero   ⊓ n      = zero
succ m ⊓ zero   = zero
succ m ⊓ succ n = succ (m ⊓ n)
\end{code}

<<maximum>>
\begin{code}
_⊔_ : ℕ → ℕ → ℕ
zero   ⊔ n      = n
succ m ⊔ zero   = succ m
succ m ⊔ succ n = succ (m ⊔ n)
\end{code}

-------------------------------------------------------------------------------

Unicode: ∸ \.-
