Feijen proof format for strict relations - this is more difficult as
the end-of-proof symbol ∎ relies on reflexibility. 

    proof
      ?
    ≡⟨ ? ⟩
    ≺⟨ ? ⟩
    ≼⟨ ? ⟩
     ?
    ∎ where open import VFP.Natural.Order.Reasoning

The format is quite flexible: it's an equality proof (x ≡ y) if only
≡-steps are used; it establishes a strict order (x ≺ y) if there is at
least one ≺-step; othwerwise it shows a non-strict order (x ≼ y).

See also
Relation.Binary.Reasoning.Preorder
https://plfa.github.io/19.08/Equality/
https://gist.github.com/UlfNorell/102f8e59744de8a69f94b39f09f70c1d

\begin{code}
{-# OPTIONS --without-K --safe #-}
--{-# OPTIONS --safe #-}

module VFP.Natural.Order.Reasoning where

open import VFP.Natural.Order.Properties public
open import VFP.Logic
open import VFP.Case

private
  variable
    m n : ℕ
  data Order : ℕ -> ℕ -> Set where
    LT : m ≺ n -> Order m n
    LE : m ≼ n -> Order m n
    EQ : m ≡ n -> Order m n

-- an anonymous module so that ⟦_⟧ and proof share the same arguments: i and j  
module _ {i j : ℕ} where
  private
    ⟦_⟧ : Order i j -> Set
    ⟦ LT _ ⟧ = i ≺ j
    ⟦ LE _ ⟧ = i ≼ j
    ⟦ EQ _ ⟧ = i ≡ j

  infix 0 proof_

  proof_ : (p : Order i j) → ⟦ p ⟧
  proof (LT p) = p
  proof (LE p) = p
  proof (EQ p) = p

infixr 1 ≺-link ≼-link ≡-link 
infix  2 _∎
syntax ≺-link x q p = x ≺⟨ p ⟩ q
syntax ≼-link x q p = x ≼⟨ p ⟩ q
syntax ≡-link x q p = x ≡⟨ p ⟩ q
-- order of last two arguments is swapped so that p's type can be inferred easily

-- the result type is x ≡ y if only ≡-links are used; it is x ≺ y if there is
-- at least one ≺-link; othwerwise it is x ≼ y

--infix 0 case_of_
--case_of_ : {A : Set} {B : Set} → A → (A → B) → B
--case x of f = f x

≺-link : (x : ℕ) {y z : ℕ} → Order y z → x ≺ y → Order x z
≺-link x (LT y≺z) x≺y  = LT (≺-transitive x≺y y≺z)
≺-link x (LE y≼z) x≺y  = LT (case y≼z of λ { (inl reflexive) -> x≺y ; (inr y≺z) -> ≺-transitive x≺y y≺z})
≺-link x (EQ y≡z) x≺y  = LT (case y≡z of λ where reflexive → x≺y)

≼-link : (x : ℕ) {y z : ℕ} → Order y z → x ≼ y → Order x z
≼-link x (LT y≺z) x≼y  = LT (case x≼y of λ { (inl reflexive) -> y≺z ; (inr x≺y) -> ≺-transitive x≺y y≺z})
≼-link x (LE y≼z) x≼y  = LE (≼-transitive x≼y y≼z)
≼-link x (EQ y≡z) x≼y  = LE (case y≡z of λ where reflexive → x≼y)

≡-link : (x : ℕ) {y z : ℕ} → Order y z → x ≡ y → Order x z
≡-link x (LT y≺z) x≡y  = LT (case x≡y of λ where reflexive -> y≺z)
≡-link x (LE y≼z) x≡y  = LE (case x≡y of λ where reflexive -> y≼z)
≡-link x (EQ y≡z) x≡y  = EQ (case x≡y of λ where reflexive -> y≡z)
-- |case| on the RHS instead of PM on the left: delayed
-- unification. We need to be able determine the "kind" of Order based
-- on the "kinds" of the arguments (so that |proof| can infer the
-- result type).

-- Ulf Norell says:
-- ^ Note: don't match on the proof here, we need to decide strict vs nonstrict for neutral proofs

_∎ : (x : ℕ) → Order x x
x ∎ = EQ reflexive

by-definition : {A : Set} {x : A} -> x ≡ x
by-definition = reflexive 
\end{code}

\begin{code}
module Examples where
  open import VFP.Natural.Properties

  -- silly example
  test : (0 + 0) ≺ 2
  test =
    proof
      0 + 0
    ≡⟨ +-left-identity ⟩
      0
    ≺⟨ n≺succ-n ⟩
      1
    ≺⟨ n≺succ-n ⟩
      2
    ≼⟨ inl reflexive ⟩
     2
    ∎
\end{code}

\begin{code}
  open import VFP.List.Properties

  decr : ∀ {elem : Set} → (x : elem) → (xs : List elem) → length (reverse xs) ≺ length (x ∷ xs)
  decr x xs = 
    proof 
      length (reverse xs) 
    ≡⟨ length-reverse xs ⟩
--    ≼⟨ inl (length-reverse xs) ⟩
      length xs 
    ≺⟨ n≺succ-n ⟩
      length (x ∷ xs)
    ∎

  weird : ∀ {elem : Set} → (bs : List elem) → Accessible (length bs) → List elem
  weird [] _ = []
  weird (x ∷ xs) (access hyp) = x ∷ weird (reverse xs) (hyp (decr x xs))
\end{code}

TODO: clean-up!

Termination proofs.
%
begin{code}
{-
  div : ℕ → ℕ → ℕ
  div (zero)  (n)     = zero
  div (suc m) (zero)  = suc m
  div (suc m) (suc n) = suc (div (suc m ∸ suc n) (suc n))
-}

  m∸n≺sm : ∀ m n → (m ∸ n) ≺ suc m
  m∸n≺sm zero zero = z≺s
  m∸n≺sm zero (suc n) = z≺s
  m∸n≺sm (suc m) zero = s≺s (n≺sn m)
  m∸n≺sm (suc m) (suc n) = ≺-transitive (m∸n≺sm m n) (s≺s (n≺sn m))

  div' : (n : ℕ) → Accessible n → ℕ → ℕ
  div' (zero)  _            (n)     = zero
  div' (suc m) _            (zero)  = suc m
  div' (suc m) (access hyp) (suc n)
    = suc (div' (suc m ∸ suc n) (hyp (m∸n≺sm m n)) (suc n))

  div : ℕ → ℕ → ℕ
  div m n = div' m (well-founded m) n
end{code}

Exercise: define one well-founded ordering in terms of another.
%
begin{code}
  open import Data.List
  
  _⪡_ : ∀ {elem : Set} → List elem → List elem → Set
  as ⪡ bs = length as ≺ length bs

  data Accessible' (elem : Set) (bs : List elem) : Set where
    access : (∀ {as} → as ⪡ bs → Accessible' elem as) → Accessible' elem bs

  go : ∀ {elem : Set} → ∀ {bs} → Accessible (length bs) → Accessible' elem bs
  go (access hyp) = access λ p → go (hyp p)

  well-founded' : ∀ {elem : Set}  → ∀ bs → Accessible' elem bs
  well-founded' bs = go (well-founded (length bs))
end{code}
