Binary search, based on non-overlapping, right-open interval trees:
    [ l , r [ ↔ l ≺ r
See also Edsger W. Dijkstra (1930–2002), Why numbering should start at zero — EWD831.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Searching.Binary-Search where

open import VFP.Natural.Midpoint.Properties-7
open import VFP.Natural.Order.Properties-7 public
open import VFP.Natural.Order.Properties using ()
                                      renaming (≺-monotone-left-+ to ≺₀-monotone-left-+;
                                                ≼-monotone-right-+ to ≼₀-monotone-right-+;
                                                zero≼n to zero≼₀n;
                                                _≺-or-≽_ to _≺-or-≽₀_;
                                                _≺?≽_ to _≺?≽₀_) public
open import VFP.Logic

private variable
  a b : ℕ
  i l m n r : ℕ
  P Q : ℕ -> Set
\end{code}

\begin{code}
infix 6 _∈_
\end{code}

-------------------------------------------------------------------------------

Recursion scheme for nested, overlapping, non-empty intervals
(„Intervallschachtelung“).

<<Interval>>
\begin{code}
data ⟦_,_⦆ : ℕ → ℕ → Set where
  Leaf  : (n  : ℕ) → ⟦ n , 1 + n ⦆
  Node  : (m  : ℕ) → ⟦ l , m ⦆ → ⟦ m , r ⦆ -> ⟦ l , r ⦆
\end{code}

The element |m| can be seen as a pivot; it is the smallest leaf
element in the *right* subtree.

NB. [_,_[ *is* _≺₃_ of Natural.Order!

<<example>>
\begin{code}
example : ⟦ 0 , 10 ⦆
\end{code}

<<0-10>>
\begin{code}
example = Node 5  (Node 2  (Node 1 (Leaf 0) (Leaf 1))
                           (Node 4 (Node 3 (Leaf 2) (Leaf 3)) (Leaf 4)))
                  (Node 8  (Node 7 (Node 6 (Leaf 5) (Leaf 6)) (Leaf 7))
                           (Node 9 (Leaf 8) (Leaf 9)))
\end{code}

This is *almost* a comparison tree à la Knuth, see TAOCP 3 p412.  The
interval tree ⟦ 0 , N ⦆ contains leafs numbered from 0 to N - 1.  The
comparison tree for N contains leafs numbered from 0 to N, see also
further below.

The elements in an interval tree are contained in its domain, see also
ordered trees.

<<domain>>
\begin{code}
domain : ⟦ l , r ⦆ -> l ≺ r
domain (Leaf n)      = one
domain (Node m t u)  = ≺-transitive (domain t) (domain u)
\end{code}

\begin{code}
yes-sir : 1 ≺ 4
yes-sir = succ (succ one)
no-no_ : ⟦ 4 , 1 ⦆ -> ⊥
no-no p = ≺-irreflexive (≺-transitive (domain p) yes-sir)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Interval search, exploring an interval tree.

<<search>>
\begin{code}
interval-search  : ⟦ l , r ⦆
                 → (oracle : ∀ n -> P n ⊎ Q n)
                 →            P l × Q r
                 → ∃ (λ i ->  P i × Q (1 + i))
interval-search (Leaf n)       oracle (p , q) = n , p , q
interval-search (Node m t u)   oracle (p , q) with oracle m
… | inr  q′  = interval-search t  oracle (p   , q′  )
… | inl  p′  = interval-search u  oracle (p′  , q   )
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

British museum search aka linear search aka sequential search. Converse of |domain|.

<<sequential>>
\begin{code}
sequential : l ≺ r -> ⟦ l , r ⦆
sequential (one)            = Leaf _
sequential (succ succ-l≺r)  = Node _ (Leaf _) (sequential succ-l≺r)
\end{code}

In other words, l ≺ r is equivalent to 

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Constructing a *balanced* interval tree — this is hard work. 

<<shrink-sig>>
\begin{code}
shrink-left   :      a ≺ b -> ⌊ a mid b ⌋ ∸ a ≺ b ∸ a
shrink-right  : 1 +  a ≺ b -> b ∸ ⌊ a mid b ⌋ ≺ b ∸ a
\end{code}

\begin{code}
shrink-left p = ∸-pseudo-monotone p (a-mid-b≺b p)
shrink-right p = ∸-pseudo-antitone (≺-transitive one p) (a≺a-mid-b p)
\end{code}

The implementation here is a variant of the one presented in the
paper. It works directly without midpoint-trees as an intermediate
step.

<<wf-binary-subdivsion-signature>>
\begin{code}
wf-binary-subdivision : (l r : ℕ) -> l ≺ r -> Accessible (r ∸ l) -> ⟦ l , r ⦆
\end{code}

<<wf-binary-subdivision-body>>
\begin{code}
wf-binary-subdivision l r (one)     (access because) = Leaf l
wf-binary-subdivision l r (succ p)  (access because) =
  let m = ⌊ l mid r ⌋ in
  Node m  (wf-binary-subdivision  l  m  (a≺a-mid-b p)          (because (shrink-left (succ p))))
          (wf-binary-subdivision  m  r  (a-mid-b≺b (succ p))  (because (shrink-right p)))
\end{code}

<<binary-subdivision>>
\begin{code}
binary-subdivision : (l r : ℕ) -> l ≺ r → ⟦ l , r ⦆
binary-subdivision l r l≺r = wf-binary-subdivision l r l≺r (well-founded (r ∸ l))
\end{code}

NB. In the recursive, we use that the interval contains at least 3
elements, so that l ≺ ⌊ l mid r ⌋ ≺ r.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Encore une fois: constructing a *balanced* interval tree. This time
the pivot is ⌈ l mid r ⌉, rather than ⌊ l mid r ⌋.

\begin{code}
shrink-left′ : succ a ≺ b -> ⌈ a mid b ⌉ ∸ a ≺ b ∸ a
shrink-left′ p = ∸-pseudo-monotone (≺-transitive one p) (⌈mid⌉-right p)

shrink-right′ : a ≺ b -> b ∸ ⌈ a mid b ⌉ ≺ b ∸ a
shrink-right′ p = ∸-pseudo-antitone p (⌈mid⌉-left p)
\end{code}

<<wf-binary-subdivision'>>
\begin{code}
wf-binary-subdivision′ : (l : ℕ) -> (r : ℕ) -> l ≺ r -> Accessible (r ∸ l) -> ⟦ l , r ⦆
wf-binary-subdivision′ l r (one)    (access because) = Leaf l
wf-binary-subdivision′ l r (succ p) (access because) =
  let m = ⌈ l mid r ⌉ in
  Node m  (wf-binary-subdivision′ l m (⌈mid⌉-left (succ p)) (because (shrink-left′ p)))
          (wf-binary-subdivision′ m r (⌈mid⌉-right p)       (because (shrink-right′ (succ p))))

binary-subdivision′ : l ≺ r -> ⟦ l , r ⦆
binary-subdivision′ {l} {r} p = wf-binary-subdivision′ l r p (well-founded (r ∸ l))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Putting things together: the (anti-) climax.

<<binary-search>>
\begin{code}
binary-search  : l ≺ r
               → (oracle : (n : ℕ) -> P n ⊎ Q n)
               →            P l × Q r
               → ∃ (λ i ->  P i × Q (1 + i))
binary-search l≺r = interval-search (binary-subdivision _ _ l≺r)
\end{code}

-------------------------------------------------------------------------------

We can strenghten the definition of interval-search: we guarantee that
the computed index is contained in the given range.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Membership.

<<membership-definition>>
\begin{code}
data _∈_ (i : ℕ) : ⟦ l , r ⦆ -> Set where
  here   : i ∈ Leaf i
  left   : {t : ⟦ l , m ⦆} -> {u : ⟦ m , r ⦆} -> i ∈ t  -> i ∈ Node m t u
  right  : {t : ⟦ l , m ⦆} -> {u : ⟦ m , r ⦆} -> i ∈ u  -> i ∈ Node m t u
\end{code}

\begin{code}
↙_ : (range : ⟦ l , r ⦆) → l ∈ range
↙ Leaf _     = here
↙ Node m t u = left (↙ t)

leftmost : (range : ⟦ l , r ⦆) → l ∈ range
leftmost range = ↙ range
\end{code}

\begin{code}
bounds : ∀ {t : ⟦ l , r ⦆} -> i ∈ t -> l ≼ i × i ≺ r
bounds (here)    = ≼-reflexive , one
bounds (left  {u = u} p) with bounds p
… | l≼i , i≺m = l≼i , ≺-transitive i≺m (domain u)
bounds (right {t = t} p) with bounds p
… | m≼i , i≺r = ≼-transitive (inr (domain t)) m≼i , i≺r

left-bound : ∀ {t : ⟦ l , r ⦆} -> i ∈ t -> l ≼ i
left-bound p = outl (bounds p)

right-bound′ : ∀ {t : ⟦ l , r ⦆} -> i ∈ t -> succ i ≼ r
right-bound′ p = ≺-is-succ-≼ (outr (bounds p))
\end{code}

\begin{code}
contains : ∀ (t : ⟦ l , r ⦆) -> l ≼ i × i ≺ r -> i ∈ t
contains (Leaf _) (p , q) rewrite the-one p q = here
contains {i = i} (Node m t u) (p , q) with m ≼?≻ i
… | inl m≼i = right (contains u (m≼i , q))
… | inr m≻i = left (contains t (p , m≻i))
\end{code}

In other words, i ∈ t is equivalent to l ≼ i × i ≺ r.

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

\begin{code}
interval-search+  : (range : ⟦ l , r ⦆)
                  → (oracle : (n : ℕ) -> P n ⊎ Q n)
                  →                       P l × Q r
                  → ∃ (λ i -> i ∈ range × P i × Q (1 + i))
interval-search+ (Leaf n)      oracle (p , q) = n , here , p , q
interval-search+ (Node m t u)  oracle (p , q) with oracle m
… | inl  p′  = let (i , i∈r , pi , qi) = interval-search+ u oracle (p′  , q   ) in i , right i∈r , pi , qi
… | inr  q′  = let (i , i∈l , pi , qi) = interval-search+ t oracle (p   , q′  ) in i , left  i∈l , pi , qi
\end{code}

\begin{code}
binary-search+  : l ≺ r
               → (oracle : (n : ℕ) -> P n ⊎ Q n)
               →                             P l × Q r
               → ∃ (λ i ->  l ≼ i × i ≺ r ×  P i × Q (1 + i))
binary-search+ l≺r oracle (p , q) with interval-search+ (binary-subdivision _ _ l≺r) oracle (p , q)
… | i , i∈r , pi , qi = let (l≼i , i≺r) = bounds i∈r in i , l≼i , i≺r , pi , qi
\end{code}

-------------------------------------------------------------------------------

Further reading:

●  §6.4.3 of “Introduction to Functional Programming”, Richard Bird and
   Philip Wadler, Prentice Hall (1988).
