+Applications of binary search.

-------------------------------------------------------------------------------

\begin{code}
{-# OPTIONS --without-K --safe #-}

module VFP.Searching.Binary-Search.Applications where

open import VFP.Searching.Binary-Search public
open import VFP.Natural.Properties public
open import VFP.Decidable public
open import VFP.Natural.Order.Interrelations using () renaming (to-≺₇ to to-≺)

private variable
  i l n r : ℕ
\end{code}

-------------------------------------------------------------------------------

Finding an even-odd pair, an instance of interval search to motivate
the set-up.

<<module-Even-Odd-Pair>>
\begin{code}
module Even-Odd-Pair where
\end{code}

<<even-odd-pair-redefined-signature>>
\begin{code}
even-odd-pair  :  ⟦ l , r ⦆
               →  (box : ℕ → ℕ) 
               →            even (box l) × odd (box r)
               →  ∃ (λ i →  even (box i) × odd (box (1 + i)))
\end{code}

<<even-odd-pair-redefined>>
\begin{code}
even-odd-pair strategy box = 
    interval-search strategy (λ n → parity (box n))
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Application: guessing a number between 0 and 9999.

\begin{code}
zero≺1000 = zero≺succ
\end{code}

<<guess-my-number>>
\begin{code}
guess-my-number : (n : ℕ) → n ≺ 1000 → ∃ (λ i → (i ≼ n) × (n ≺ 1 + i))
guess-my-number n n≺1000 = 
    binary-search zero≺1000 (λ i → i ≼?≻ n) (zero≼n n , n≺1000)
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Application: integer/natural square root. (More generally: the
pseudo-inverse of a function.)

<<square>>
\begin{code}
_² : ℕ → ℕ
(n)² = n * n
\end{code}

\begin{code}
square-lemma′ : ∀ n → n ≺₀ (1 + n)²
square-lemma′ n =
    (proof
      n
    ≡⟨ symmetric +-right-identity ⟩
      n + 0
    ≼⟨ ≼₀-monotone-right-+ (zero≼₀n (n * succ n)) ⟩
      n + n * succ n
    ≺⟨ ≺₀-monotone-left-+ (n≺₀succ-n {n = n}) ⟩
      succ n + n * succ n
    ≡⟨ reflexive ⟩
      (1 + n)²
    ∎) where open import VFP.Natural.Order.Reasoning renaming (n≺succ-n to n≺₀succ-n)
\end{code}

\begin{code}
square-lemma : ∀ n → n ≺ (1 + n)²
square-lemma n = to-≺ (square-lemma′ n)
\end{code}

NO, one will do:

<<square-root>>
\begin{code}
square-root : (n : ℕ) → ∃ (λ i → i ² ≼ n × n ≺ (1 + i)²)
square-root n = binary-search  (zero≺succ-n n) (λ i → i ² ≼?≻ n)
                               (zero≼n n , square-lemma n)
\end{code}

\begin{code}
⌊√_⌋ : ℕ → ℕ
⌊√ n ⌋ = outl (square-root n)
\end{code}

Self-certifying code.

Ctrl-C Ctrl-N
  ⌊√ 3 ⌋
  ⌊√ 4 ⌋
  outl ⌊√ 4711 ⌋

Indirect definition of square root: ⌊√_⌋ is right adjoint to _².

begin{code}
gc : ∀ n x → (n)² ≼′ x  ←  n ≼′ ⌊√ x ⌋
gc = {!!}
end{code}

Asymmetric definition needed as we cannot (?) show that ⌊√_⌋ is
monotone.

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Application: Beat Your Neighbours!

see also, https://boardgamegeek.com/boardgame/36833/beat-your-neighbours

<<xbeat-your-neighbours>>
\begin{code}
beat-your-neighbours-bin  :  (box : ℕ → ℕ) 
                           →  l ≺ r
                           →            box l ≼ box (1 + l) × box r ≽ box (1 + r)
                           →  ∃ (λ i →  box i ≼ box (1 + i) × box (1 + i) ≽ box (2 + i))
beat-your-neighbours-bin box l≺r =
  binary-search l≺r (λ i → box i ≼?≽ box (1 + i))
\end{code}

Self-certifying code.

<<beat-your-neighbours>>
\begin{code}
beat-your-neighbours  :  (box : ℕ → ℕ) 
                       →  ⟦ l , r ⦆
                       →            box l ≼ box (1 + l) × box r ≽ box (1 + r)
                       →  ∃ (λ i →  box i ≼ box (1 + i) × box (1 + i) ≽ box (2 + i))
beat-your-neighbours box strategy =
  interval-search strategy (λ i → box i ≼?≽ box (1 + i))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Boundary boxes are 0, not -∞.

\begin{code}
module BeatYourNeighbours where
  open import VFP.List

  boxes : List ℕ
  boxes = 4 ∷ 2 ∷ 7 ∷ 6 ∷ 5 ∷ 3 ∷ 9 ∷ 8 ∷ 1 ∷ []

  box : ℕ → ℕ
  box (zero)      = 0
  box (succ n) with n ≺?≽₀ length boxes
  … | inl n≺len = lookup boxes n≺len
  … | inr n≽len = 0

  test = beat-your-neighbours-bin {r = length boxes} box zero≺succ (zero≼n _ , zero≼n _)
\end{code}

Ctrl-C Ctrl-N
  BeatYourNeighbours.test

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Bird&Wadler: 

\begin{code}
private variable
  P : ℕ → Set
\end{code}

<<find-interface>>
\begin{code}
find  : l ≺ r
      → (monotone : ∀ {a b} → a ≼ b → (P a → P b))
      → (decide : ∀ n -> ¬ P n ⊎ P n)
      → ¬ P l × P r
      → ∃ (λ j ->  P j × (∀ i → i ≺ j → ¬ P i))
\end{code}

<<find-implementation>>
\begin{code}
find l≺r monotone decide pre with binary-search l≺r decide pre
… |  j , ¬P-j , P-1+j 
  =  1 + j , P-1+j ,  λ i (i≺1+j : i ≺ 1 + j) P-i →
                        ¬P-j (monotone (≺-succ-is-≼ i≺1+j) P-i)
\end{code}

… | j , ¬P-j , P-1+j =  1 + j , P-1+j ,
  λ i (i≺1+j : i ≺ 1 + j) P-i → ¬P-j (monotone (≺-succ-is-≼ i≺1+j) P-i)

Uniqueness, if P is monotone.

\begin{code}
unique  : (monotone : ∀ {i j} → i ≼ j → (P i → P j))
        → (i : ℕ) → ¬ P i × P (1 + i)
        → (j : ℕ) → ¬ P j × P (1 + j)
        → i ≡ j
unique monotone i (¬P-i , P-1+i) j (¬P-j , P-1+j) with i ≼?≻ j
... | inl (inr  i≺j)  = ex-falso-quodlibet (¬P-j (monotone (≺-is-succ-≼ i≺j) P-1+i))
... | inl (inl  i≡j)  = i≡j
... | inr       i≻j   = ex-falso-quodlibet (¬P-i (monotone (≺-is-succ-≼ i≻j) P-1+j))
\end{code}


A variant of find where no preconditions are given and therefore we
cannot guarantee the existence of a smallest index satisfying
P. Instead, we *decide* whether there is such an index in the
interval.

<<find>>
\begin{code}
find?  :  ∀ l r
      →  (t : ⟦ l , r ⦆)
      →  (monotone : ∀ {i j} → i ≼ j → (P i → P j))
      →  (decide : ∀ n → ¬ P n ⊎ P n)
      →  Decide (∃ (λ i → i ∈ t × ¬ P i × P (1 + i)))
find? l r t monotone decide with decide l | decide r
… | inl ¬pl  | inl ¬pr  = no   (λ { (i , i∈t , ¬p , p′) → ¬pr (monotone (right-bound′ i∈t) p′) })
… | inl ¬pl  | inr  pr  = yes  (interval-search+ t decide (¬pl , pr))
… | inr  pl  | inl ¬pr  = no   (λ _ → ¬pr (monotone (inr (domain t)) pl))
… | inr  pl  | inr  pr  = no   (λ { (i , i∈t , ¬p , p′) → ¬p (monotone (left-bound i∈t) pl) })
\end{code}

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Searching a table.

<<table-lookup>>
\begin{code}
table-lookup  : l ≺ r
              → (table : ℕ → ℕ) 
              → (key : ℕ)
              →           table l ≼ key × key ≺ table r
              → ∃ (λ i →  table i ≼ key × key ≺ table (1 + i))
table-lookup l≺r table key = binary-search l≺r (λ i → table i ≼?≻ key)
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

The classic application: searching an ordered table: we *decide*
whether there is an i with box i ≡ key.

Assumes: box l ≼ key × key ≺ box r. 

<<ord-table-lookup-interface>>
\begin{code}
ord-table-lookup  : l ≺ r
                  → (table : ℕ → ℕ) 
                  → (monotone : ∀ {a b} → a ≼ b → table a ≼ table b)
                  → (key : ℕ)
                  → table l ≼ key × key ≺ table r
                  → Decide (∃ (λ i → table i ≡ key))
\end{code}

<<ord-table-lookup-implementation-1>>
\begin{code}
ord-table-lookup l≺r table monotone key pre
  with table-lookup l≺r table key pre
... | i , inl  pᵢ , qᵢ  = yes  (i , pᵢ)
... | i , inr  pᵢ , qᵢ  = no   key-not-found
  where  key-not-found : ¬ ∃ (λ j → table j ≡ key)
\end{code}
<<ord-table-lookup-implementation-2>>
\begin{code}
         key-not-found (j , reflexive) with j ≼?≻ i
         ... | inl  j≼i  = ≺-irreflexive (≺-then-≼ pᵢ (monotone j≼i))
         ... | inr  j≻i  = ≺-irreflexive (≺-then-≼ qᵢ (monotone (≺-is-succ-≼ j≻i)))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Assumes: box l ≼ key × key ≼ box r. This allows, for example, all boxes
to contain the same number.

\begin{code}
nonstrict-ord-table-lookup  : l ≺ r
                            → (box : ℕ → ℕ) 
                            → (monotone : ∀ {i j} → i ≼ j → box i ≼ box j)
                            → (key : ℕ)
                            → box l ≼ key × key ≼ box r
                            → Decide (∃ (λ i → box i ≡ key))
nonstrict-ord-table-lookup l≺r box monotone key (p , q)
  with binary-search l≺r (λ m → box m ≼?≽ key) (p , q)
… | i , inl reflexive , qᵢ = yes (i , reflexive)
… | i , inr pᵢ , inl reflexive = yes (succ i , reflexive)
… | i , inr pᵢ , inr qᵢ = no not-found
  where not-found : ¬ ∃ (λ j → box j ≡ key)
        not-found (j , reflexive) with i ≺?≽ j
        … | inl  i≺j  = ≺-irreflexive (≺-then-≼ qᵢ (monotone (≺-is-succ-≼ i≺j)))
        … | inr  i≽j  = ≺-irreflexive (≺-then-≼ pᵢ (monotone i≽j))
\end{code}

NB. Monotonicity is only used in the negative case. BTW Knuth assumes
a *strict* ordering (as he builds on a ternary/three-way comparison).

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

Uses interval-search+ which guarantees that the index is in the given
range. The negative part is weaker (there could be an index i outside
the interval satisfying box i ≡ key); the positive is stronger.

\begin{code}
table-lookup+ : l ≺ r
              → (box : ℕ → ℕ) 
              → (monotone : ∀ {i j} → i ≼ j → box i ≼ box j)
              → (key : ℕ)
              → box l ≼ key × key ≺ box r
              → Decide (∃ (λ i → l ≼ i × i ≺ r × box i ≡ key))
table-lookup+ l≺r box monotone key (p , q)
  with binary-search+ l≺r (λ m → box m ≼?≻ key) (p , q)
… | i , l≼i , i≺r , inl reflexive , qᵢ = yes (i , l≼i , i≺r , reflexive)
… | i , l≼i , i≺r , inr pᵢ        , qᵢ = no not-found
  where not-found : ¬ ∃ (λ j → l ≼ j × j ≺ r × box j ≡ key)
        not-found (j , l≼i , i≺r , reflexive) with i ≺?≽ j
        … | inl i≺j = ≺-irreflexive (≺-then-≼ qᵢ (monotone (≺-is-succ-≼ i≺j)))
        … | inr i≽j = ≺-irreflexive (≺-then-≼ pᵢ (monotone i≽j))
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

table-lookup+ with strict monotonicity and no preconditions.

\begin{code}
weaken : {box : ℕ → ℕ}
       → (∀ {i j} → i ≺ j → box i ≺ box j)
       → (∀ {i j} → i ≼ j → box i ≼ box j)
weaken monotone (inl reflexive) = inl reflexive
weaken monotone (inr x) = inr (monotone x)
\end{code}

\begin{code}
table-lookup+′ : (l r : ℕ)
               → l ≺ r
               → (box : ℕ → ℕ) 
               → (monotone : ∀ {i j} → i ≺ j → box i ≺ box j)
               → (key : ℕ)
               → Decide (∃ (λ i → l ≼ i × i ≺ r × box i ≡ key))
table-lookup+′ l r l≺r box monotone key with box l ≼?≻ key | key ≺?≽ box r
… | inl xx | inl x = table-lookup+ l≺r box (weaken monotone) key (xx , x)
… | inl _ | inr  box-r≼key = no λ { (i , l≼i , i≺r , reflexive) → ≺-irreflexive (≺-then-≼ (monotone i≺r) box-r≼key) }  -- here strict monotoneity is required
… | inr key≺box-l | _ = no (λ {(i , l≼i , i≺r , reflexive) → ≺-irreflexive (≺-then-≼ key≺box-l (weaken monotone l≼i))})
\end{code}

table-lookup+ without preconditions. To reduce clutter, we allow the
index to be equal to r.

\begin{code}
table-lookup+″ : (l r : ℕ)
               → l ≺ r
               → (box : ℕ → ℕ) 
               → (monotone : ∀ {i j} → i ≼ j → box i ≼ box j)
               → (key : ℕ)
               → Decide (∃ (λ i → l ≼ i × i ≼ r × box i ≡ key))
table-lookup+″ l r l≺r box monotone key with box l ≼?≻ key | key ≺?≽ box r
table-lookup+″ l r l≺r box monotone key | inl xx | inl x with table-lookup+ l≺r box monotone key (xx , x)
… | yes (i , l≼i , i≺r , reflexive) = yes (i , l≼i , inr i≺r , reflexive)
… | no ¬f = no λ where
                   (i , l≼i , inl reflexive , reflexive) → ≺-irreflexive x
                   (i , l≼i , inr x , reflexive) → ¬f (i , l≼i , x , reflexive)
table-lookup+″ l r l≺r box monotone key | inl _ | inr (inl x) = yes (r , inr l≺r , inl reflexive , x)
table-lookup+″ l r l≺r box monotone key | inl _ | inr (inr box-r≺key) = no λ { (i , l≼i , i≼r , reflexive) → ≺-irreflexive (≺-then-≼ box-r≺key (monotone i≼r)) }
table-lookup+″ l r l≺r box monotone key | inr key≺box-l | _ = no (λ {(i , l≼i , i≼r , reflexive) → ≺-irreflexive (≺-then-≼ key≺box-l (monotone l≼i))})
\end{code}

. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
-------------------------------------------------------------------------------
